# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

from . import errors as errors
from . import ops as ops
from . import project as project
from .components.annotation import (
    NO_GROUP,
    Annotation,
    AnnotationType,
    Bbox,
    BinaryMaskImage,
    Caption,
    Categories,
    Colormap,
    CompiledMask,
    CompiledMaskImage,
    Cuboid3d,
    Ellipse,
    IndexMaskImage,
    Label,
    LabelCategories,
    Mask,
    MaskCategories,
    Points,
    PointsCategories,
    Polygon,
    PolyLine,
    RgbColor,
    RleMask,
)
from .components.cli_plugin import CliPlugin
from .components.dataset import Dataset, DatasetPatch, DatasetSubset, ItemStatus, eager_mode
from .components.dataset_base import (
    DEFAULT_SUBSET_NAME,
    CategoriesInfo,
    DatasetBase,
    DatasetItem,
    IDataset,
    SubsetBase,
)
from .components.environment import Environment, PluginRegistry
from .components.exporter import Exporter, ExportErrorPolicy, FailingExportErrorPolicy
from .components.hl_ops import (  # pylint: disable=redefined-builtin
    export,
    filter,
    merge,
    run_model,
    transform,
    validate,
)
from .components.importer import FailingImportErrorPolicy, Importer, ImportErrorPolicy
from .components.launcher import Launcher, ModelTransform
from .components.media import ByteImage, Image, MediaElement, Video, VideoFrame
from .components.media_manager import MediaManager
from .components.progress_reporting import NullProgressReporter, ProgressReporter
from .components.transformer import ItemTransform, Transform
from .components.validator import Validator
from .components.visualizer import Visualizer
from .version import VERSION



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import struct
from io import BufferedReader
from typing import Optional

from datumaro.components.crypter import NULL_CRYPTER, Crypter
from datumaro.components.errors import DatasetImportError
from datumaro.components.media import Image, MediaElement, MediaType, PointCloud
from datumaro.plugins.data_formats.datumaro_binary.format import DatumaroBinaryPath
from datumaro.plugins.data_formats.datumaro_binary.mapper import DictMapper
from datumaro.plugins.data_formats.datumaro_binary.mapper.dataset_item import DatasetItemMapper

from ..datumaro.base import DatumaroBase


class DatumaroBinaryBase(DatumaroBase):
    """"""

    def __init__(self, path: str, encryption_key: Optional[bytes] = None):
        self._fp: Optional[BufferedReader] = None
        self._crypter = Crypter(encryption_key) if encryption_key is not None else NULL_CRYPTER
        super().__init__(path)

    def _get_dm_format_version(self, path: str) -> str:
        with open(path, "rb") as fp:
            self._fp = fp
            self._check_signature()
            dm_format_version = self._read_version()
        return dm_format_version

    def _load_impl(self, path: str) -> None:
        """Actual implementation of loading Datumaro binary format."""
        try:
            with open(path, "rb") as fp:
                self._fp = fp
                self._check_signature()
                self._read_version()
                self._check_encryption_field()
                self._read_info()
                self._read_categories()
                self._read_media_type()
                self._read_items()
        finally:
            self._fp = None

    def _check_signature(self):
        signature = self._fp.read(DatumaroBinaryPath.SIGNATURE_LEN).decode()
        DatumaroBinaryPath.check_signature(signature)

    def _check_encryption_field(self):
        len_byte = self._fp.read(4)
        _bytes = self._fp.read(struct.unpack("I", len_byte)[0])

        if not self._crypter.handshake(_bytes):
            raise DatasetImportError("Encryption key handshake fails. You give a wrong key.")

    def _read_header(self, use_crypter: bool = True):
        len_byte = self._fp.read(4)
        _bytes = self._fp.read(struct.unpack("I", len_byte)[0])
        if use_crypter:
            _bytes = self._crypter.decrypt(_bytes)
        header, _ = DictMapper.backward(_bytes)
        return header

    def _read_version(self) -> str:
        return self._read_header(use_crypter=False)["dm_format_version"]

    def _read_info(self):
        self._infos = self._read_header()

    def _read_categories(self):
        categories = self._read_header()
        self._categories = self._load_categories({"categories": categories})

    def _read_media_type(self):
        media_type = self._read_header()["media_type"]
        if media_type == MediaType.IMAGE:
            self._media_type = Image
        elif media_type == MediaType.POINT_CLOUD:
            self._media_type = PointCloud
        elif media_type == MediaType.MEDIA_ELEMENT:
            self._media_type = MediaElement
        else:
            raise NotImplementedError(f"media_type={media_type} is currently not supported.")

    def _read_items(self):
        (n_items,) = struct.unpack("I", self._fp.read(4))
        offset = 0
        _bytes = self._crypter.decrypt(self._fp.read())

        self._items = []

        for _ in range(n_items):
            item, offset = DatasetItemMapper.backward(_bytes, offset)
            if item.media is not None:
                item.media.set_crypter(self._crypter)
            self._items.append(item)



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from enum import Enum, auto
from typing import Dict, List

from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset import IDataset


class Severity(Enum):
    warning = auto()
    error = auto()


class TaskType(Enum):
    classification = auto()
    detection = auto()
    segmentation = auto()


class Validator(CliPlugin):
    def validate(self, dataset: IDataset) -> Dict:
        """
        Returns the validation results of a dataset based on task type.

        Args:
            dataset (IDataset): Dataset to be validated

        Raises:
            ValueError

        Returns:
            validation_results (dict):
                Dict with validation statistics, reports and summary.
        """

        validation_results = {}
        if not isinstance(dataset, IDataset):
            raise TypeError("Invalid dataset type '%s'" % type(dataset))

        # generate statistics
        stats = self.compute_statistics(dataset)
        validation_results["statistics"] = stats

        # generate validation reports and summary
        reports = self.generate_reports(stats)
        reports = list(map(lambda r: r.to_dict(), reports))

        summary = {
            "errors": sum(map(lambda r: r["severity"] == "error", reports)),
            "warnings": sum(map(lambda r: r["severity"] == "warning", reports)),
        }

        validation_results["validation_reports"] = reports
        validation_results["summary"] = summary

        return validation_results

    def compute_statistics(self, dataset: IDataset) -> Dict:
        """
        Computes statistics of the dataset based on task type.

        Args:
            dataset (IDataset): a dataset to be validated

        Returns:
            stats (dict): A dict object containing statistics of the dataset.
        """
        raise NotImplementedError("Must be implemented in a subclass")

    def generate_reports(self, stats: Dict) -> List[Dict]:
        raise NotImplementedError("Must be implemented in a subclass")



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from enum import Enum, auto

import cv2
import numpy as np
from scipy.linalg import orth

from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME
from datumaro.components.transformer import Transform
from datumaro.util import parse_str_enum_value


class Algorithm(Enum):
    gradient = auto()
    # other algorithms will be added


class OverSamplingMethod(Enum):
    random = auto()
    similarity = auto()


class UnderSamplingMethod(Enum):
    uniform = auto()
    inverse = auto()


class NDR(Transform, CliPlugin):
    """
    Removes near-duplicated images in subset|n
    |n
    Remove duplicated images from a dataset. Keep at most `-k/--num_cut`
    resulting images.|n
    |n
    Available oversampling policies (the `-e` parameter):|n
    |s|s- `random` - sample from removed data randomly|n
    |s|s- `similarity` - sample from removed data with ascending similarity score|n
    |n
    Available undersampling policies (the `-u` parameter):|n
    |s|s- `uniform` - sample data with uniform distribution|n
    |s|s- `inverse` - sample data with reciprocal of the number of number of
    items with the same similarity|n
    |n
    Example: apply NDR, return no more than 100 images|n

    .. code-block::

    |s|s%(prog)s|n
    |s|s|s|s--working_subset train|n
    |s|s|s|s--algorithm gradient|n
    |s|s|s|s--num_cut 100|n
    |s|s|s|s--over_sample random|n
    |s|s|s|s--under_sample uniform
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-w",
            "--working_subset",
            default=None,
            help="Name of the subset to operate (default: %(default)s)",
        )
        parser.add_argument(
            "-d",
            "--duplicated_subset",
            default="duplicated",
            help="Name of the subset for the removed data " "after NDR runs (default: %(default)s)",
        )
        parser.add_argument(
            "-a",
            "--algorithm",
            default=Algorithm.gradient.name,
            choices=[algo.name for algo in Algorithm],
            help="Name of the algorithm to use (default: %(default)s)",
        )
        parser.add_argument(
            "-k", "--num_cut", default=None, type=int, help="Maximum output dataset size"
        )
        parser.add_argument(
            "-e",
            "--over_sample",
            default=OverSamplingMethod.random.name,
            choices=[method.name for method in OverSamplingMethod],
            help="The policy to use when num_cut is bigger "
            "than result length (default: %(default)s)",
        )
        parser.add_argument(
            "-u",
            "--under_sample",
            default=UnderSamplingMethod.uniform.name,
            choices=[method.name for method in UnderSamplingMethod],
            help="The policy to use when num_cut is smaller "
            "than result length (default: %(default)s)",
        )
        parser.add_argument("-s", "--seed", type=int, help="Random seed")
        return parser

    def __init__(
        self,
        extractor,
        working_subset,
        duplicated_subset="duplicated",
        algorithm=None,
        num_cut=None,
        over_sample=None,
        under_sample=None,
        seed=None,
        **kwargs,
    ):
        """
        Near-duplicated image removal

        Arguments
        ---------
        working_subset: str
            name of the subset to operate
            if None, use DEFAULT_SUBSET_NAME
        duplicated_subset: str
            name of the subset for the removed data after NDR runs
        algorithm: str
            name of the algorithm to use
            "gradient" only for now.
        num_cut: int
            number of outputs you want.
            the algorithm will cut whole dataset to this amount
            if None, return result without any modification
        over_sample: "random" or "similarity"
            specify the strategy when num_cut > length of the result after removal
            if random, sample from removed data randomly
            if similarity, select from removed data with ascending
            order of similarity
        under_sample: "uniform" or "inverse"
            specify the strategy when num_cut < length of the result after removal
            if uniform, sample data with uniform distribution
            if inverse, sample data with reciprocal of the number
            of data which have same hash key

        Algorithm Specific for gradient
            block_shape: tuple, (h, w)
                for the robustness, this function will operate on blocks
                mean and variance will be calculated on this block
            hash_dim: int
                dimension(or bit) of the hash function
            sim_threshold: float
                the threshold value for saving hash-collided samples.
                larger value means more generous, i.e., saving more samples
        Return
        ---------------
        None, other subsets combined with the result
        """
        super().__init__(extractor)

        if not working_subset:
            working_subset = DEFAULT_SUBSET_NAME
        if working_subset not in extractor.subsets():
            raise ValueError("Invalid working_subset name")
        self.working_subset = working_subset

        # parameter validation before main runs
        if working_subset == duplicated_subset:
            raise ValueError("working_subset == duplicated_subset")

        algorithm = parse_str_enum_value(
            algorithm,
            Algorithm,
            default=Algorithm.gradient,
            unknown_member_error="Unknown algorithm '{value}'.",
        )
        over_sample = parse_str_enum_value(
            over_sample,
            OverSamplingMethod,
            default=OverSamplingMethod.random,
            unknown_member_error="Unknown oversampling method '{value}'.",
        )
        under_sample = parse_str_enum_value(
            under_sample,
            UnderSamplingMethod,
            default=UnderSamplingMethod.uniform,
            unknown_member_error="Unknown undersampling method '{value}'.",
        )

        if seed:
            self.seed = seed
        else:
            self.seed = None
        self.working_subset = working_subset
        self.duplicated_subset = duplicated_subset
        self.algorithm = algorithm

        self.num_cut = num_cut
        self.over_sample = over_sample
        self.under_sample = under_sample

        self.algorithm_specific = kwargs
        self._initialized = False

    def _remove(self):
        if self.seed:
            np.random.seed(self.seed)

        working_subset = self._extractor.get_subset(self.working_subset)
        having_image = []
        all_imgs = []
        for item in working_subset:
            if item.media.has_data:
                having_image.append(item)
                img = item.media.data
                # Not handle empty image, as utils/image.py if check empty
                if len(img.shape) == 2:
                    img = np.stack((img,) * 3, axis=-1)
                elif len(img.shape) == 3:
                    if img.shape[2] == 1:
                        img = np.stack((img[:, :, 0],) * 3, axis=-1)
                    elif img.shape[2] == 4:
                        img = img[..., :3]
                    elif img.shape[2] == 3:
                        pass
                    else:
                        raise ValueError(
                            "Item %s: invalid image shape: "
                            "unexpected number of channels (%s)" % (item.id, img.shape[2])
                        )
                else:
                    raise ValueError(
                        "Item %s: invalid image shape: "
                        "unexpected number of dimensions (%s)" % (item.id, len(img.shape))
                    )

                if self.algorithm == Algorithm.gradient:
                    # Calculate gradient
                    img = self._cgrad_feature(img)
                else:
                    raise NotImplementedError()
                all_imgs.append(img)
            else:
                log.debug("Skipping item %s: no image data available", item.id)

        if self.num_cut and self.num_cut > len(all_imgs):
            raise ValueError("The number of images is smaller than the cut you want")

        if self.algorithm == Algorithm.gradient:
            all_key, fidx, kept_index, key_counter, removed_index_with_sim = self._gradient_based(
                all_imgs, **self.algorithm_specific
            )
        else:
            raise NotImplementedError()

        kept_index = self._keep_cut(
            self.num_cut,
            all_key,
            fidx,
            kept_index,
            key_counter,
            removed_index_with_sim,
            self.over_sample,
            self.under_sample,
        )
        self.kept_item_id = set(having_image[ii].id for ii in kept_index)

    def _gradient_based(self, all_imgs, block_shape=(4, 4), hash_dim=32, sim_threshold=0.5):
        if len(block_shape) != 2:
            raise ValueError("Invalid block_shape")
        if block_shape[0] <= 0 or block_shape[1] <= 0:
            raise ValueError("block_shape should be positive")
        if sim_threshold <= 0:
            raise ValueError("sim_threshold should be large than 0")
        if hash_dim > 3 * block_shape[0] * block_shape[1]:
            raise ValueError("hash_dim should be smaller than feature shape")
        if hash_dim <= 0:
            raise ValueError("hash_dim should be positive")

        # Compute hash keys from all the features
        all_clr = np.reshape(np.array(all_imgs), (len(all_imgs), -1))
        all_key = self._project(all_clr, hash_dim)

        # Remove duplication using hash
        clr_dict = {}
        key_counter = {}
        kept_index = []
        removed_index_with_similarity = dict()

        fidx = np.random.permutation(np.arange(len(all_imgs)))
        for ii in fidx:
            key = all_key[ii]
            clr = all_clr[ii]
            if key not in clr_dict:
                clr_dict[key] = [clr]
                key_counter[key] = 1
                kept_index.append(ii)
                continue

            # Hash collision: compare dot-product based feature similarity
            # the value for maximizing the gap
            # between duplicated and non-duplicated
            large_exponent = 50
            max_sim = np.max(np.dot(clr_dict[key], clr) ** large_exponent)

            # Keep if not a duplicated one
            if max_sim < sim_threshold:
                clr_dict[key].append(clr)
                key_counter[key] += 1
                kept_index.append(ii)
            else:
                removed_index_with_similarity[ii] = max_sim
        return all_key, fidx, kept_index, key_counter, removed_index_with_similarity

    def _keep_cut(
        self,
        num_cut,
        all_key,
        fidx,
        kept_index,
        key_counter,
        removed_index_with_similarity,
        over_sample,
        under_sample,
    ):
        if num_cut and num_cut > len(kept_index):
            if over_sample == OverSamplingMethod.random:
                selected_index = np.random.choice(
                    list(set(fidx) - set(kept_index)), size=num_cut - len(kept_index), replace=False
                )
            elif over_sample == OverSamplingMethod.similarity:
                removed_index_with_similarity = [
                    [key, value] for key, value in removed_index_with_similarity.items()
                ]
                removed_index_with_similarity.sort(key=lambda x: x[1])
                selected_index = [
                    index for index, _ in removed_index_with_similarity[: num_cut - len(kept_index)]
                ]
            kept_index.extend(selected_index)
        elif num_cut and num_cut < len(kept_index):
            if under_sample == UnderSamplingMethod.uniform:
                prob = None
            elif under_sample == UnderSamplingMethod.inverse:
                # if inverse - probability with inverse
                # of the collision(number of same hash key)
                # [x1, x2, y1, y2, y3, y4, z1, z2, z3]. x, y and z for hash key
                # i.e. there are 4 elements which have hash key y.
                # then the occurrence will be [2, 4, 3] and reverse of them
                # will be [1/2, 1/4, 1/3]
                # Normalizing them by dividing with sum, we get [6/13, 3/13, 4/13]
                # Then the key x will be sampled with probability 6/13
                # and each point, x1 and x2, will share same prob. 3/13
                key_with_reverse_occur = {key: 1 / key_counter[key] for key in key_counter}
                reverse_occur_sum = sum(key_with_reverse_occur.values())
                key_normalized_reverse_occur = {
                    key: reverse_occur / reverse_occur_sum
                    for key, reverse_occur in key_with_reverse_occur.items()
                }
                prob = [
                    key_normalized_reverse_occur[all_key[ii]] / key_counter[all_key[ii]]
                    for ii in kept_index
                ]
            kept_index = np.random.choice(kept_index, size=num_cut, replace=False, p=prob)

        return kept_index

    @staticmethod
    def _cgrad_feature(img, out_wh=(8, 8)):
        if img.dtype == "uint8":
            img = img.astype(float) / 255.0
        else:
            img = img.astype(float)

        r_img = cv2.resize(img, out_wh, interpolation=cv2.INTER_AREA)
        r2 = cv2.resize(img**2, out_wh, interpolation=cv2.INTER_AREA)

        r2 -= r_img**2
        r2 = np.sqrt(np.maximum(r2, 0))

        # mean and variance feature, zero padding for gradient computation
        rr = np.pad(np.concatenate([r_img, r2], axis=-1), ((1, 1), (1, 1), (0, 0)))

        # compute gradients along x- and y-axes
        rx = rr[1:-1, :-2, :] - rr[1:-1, 2:, :]
        ry = rr[:-2, 1:-1, :] - rr[2:, 1:-1, :]

        # concat and l2 normalize
        res = np.concatenate([rx, ry], axis=-1)
        res = res / np.sqrt(np.sum(res**2))
        return res

    @staticmethod
    def _project(feat, hash_dim=32):
        """
        Project feat to hash_dim space and create hexadecimal string key

        Arguments
        ------------
        feat : ndarray
            feature to project and hash
        hash_dim : int
            specified dimension of the hashed output
        feature dimension should larger than hash_dim
        """
        proj = None
        ndim = feat.shape[-1]
        feat = np.reshape(feat, (-1, ndim))
        # random projection matrix would become unstable, so reject such cases
        assert ndim >= hash_dim, "{} is smaller than hash_dim({})".format(ndim, hash_dim)

        # compute the random projection matrix
        for _ in range(100):
            # try to get an orthonormal projection matrix
            proj = orth(np.random.uniform(-1, 1, (ndim, ndim)))[:, :hash_dim]
            if proj.shape[1] == hash_dim:
                break
        if proj is None:
            # if failed to get an orthonormal one, just use a random one instead
            proj = np.random.uniform(-1, 1, (ndim, ndim))
            proj /= np.sqrt(np.sum(proj**2, axis=1, keepdims=True))

        # simple binarization
        # compute dot product between each feature and each projection basis,
        # then use its sign for the binarization
        feat_binary = np.dot(feat, proj) >= 0

        # generate hash key strings
        # assign hex string from each consecutive 16 bits and concatenate
        _all_key = np.packbits(feat_binary, axis=-1)
        _all_key = np.array(
            list(map(lambda row: "".join(["{:02x}".format(r) for r in row]), _all_key))
        )
        if len(_all_key) == 1:
            return _all_key[0]
        else:
            return _all_key

    def _check_subset(self, item):
        if item.subset:
            if item.subset == self.working_subset:
                if item.id in self.kept_item_id:
                    return item.subset
                else:
                    return self.duplicated_subset
            else:
                return item.subset
        else:
            return DEFAULT_SUBSET_NAME

    def __iter__(self):
        if not self._initialized:
            self._remove()
            self._initialized = True
        for item in self._extractor:
            yield self.wrap_item(item, subset=self._check_subset(item))



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp

from datumaro.components.annotation import (
    AnnotationType,
    Bbox,
    Caption,
    Cuboid3d,
    Ellipse,
    Label,
    LabelCategories,
    MaskCategories,
    Points,
    PointsCategories,
    Polygon,
    PolyLine,
    RleMask,
)
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import DatasetImportError
from datumaro.components.media import Image, MediaElement, PointCloud
from datumaro.util import parse_json_file
from datumaro.version import VERSION

from .format import DATUMARO_FORMAT_VERSION, DatumaroPath


class DatumaroBase(SubsetBase):
    LEGACY_VERSION = "legacy"
    CURRENT_DATUMARO_FORMAT_VERSION = DATUMARO_FORMAT_VERSION

    # If Datumaro format version goes up, it will be
    # ALLOWED_VERSIONS = {LEGACY_VERSION, 1.0, ..., CURRENT_DATUMARO_FORMAT_VERSION}
    ALLOWED_VERSIONS = {LEGACY_VERSION, CURRENT_DATUMARO_FORMAT_VERSION}

    def __init__(self, path):
        assert osp.isfile(path), path

        dm_version = self._get_dm_format_version(path)

        # when backward compatibility happen, we should implement version specific readers
        if dm_version not in self.ALLOWED_VERSIONS:
            raise DatasetImportError(
                f"Datumaro format version of the given dataset is {dm_version}, "
                f"but not supported by this Datumaro version: {VERSION}. "
                f"The allowed datumaro format versions are {self.ALLOWED_VERSIONS}. "
                "Please install the latest Datumaro."
            )

        self.default_reader(path=path)

    def default_reader(self, path: str):
        """
        Default Datumaro reader for the latest version
        """
        rootpath = ""
        if path.endswith(osp.join(DatumaroPath.ANNOTATIONS_DIR, osp.basename(path))):
            rootpath = path.rsplit(DatumaroPath.ANNOTATIONS_DIR, maxsplit=1)[0]

        images_dir = ""
        if rootpath and osp.isdir(osp.join(rootpath, DatumaroPath.IMAGES_DIR)):
            images_dir = osp.join(rootpath, DatumaroPath.IMAGES_DIR)
        self._images_dir = images_dir

        pcd_dir = ""
        if rootpath and osp.isdir(osp.join(rootpath, DatumaroPath.PCD_DIR)):
            pcd_dir = osp.join(rootpath, DatumaroPath.PCD_DIR)
        self._pcd_dir = pcd_dir

        related_images_dir = ""
        if rootpath and osp.isdir(osp.join(rootpath, DatumaroPath.RELATED_IMAGES_DIR)):
            related_images_dir = osp.join(rootpath, DatumaroPath.RELATED_IMAGES_DIR)
        self._related_images_dir = related_images_dir

        super().__init__(subset=osp.splitext(osp.basename(path))[0])
        self._load_impl(path)

    def _get_dm_format_version(self, path: str):
        """
        Get Datumaro format at exporting the dataset

        Note that the regacy Datumaro doesn't store the version into exported dataset.
        Thus it returns DatumaroBase.REGACY_VERSION
        """
        self._parsed_anns = parse_json_file(path)
        return self._parsed_anns.get("dm_format_version", self.LEGACY_VERSION)

    def _load_impl(self, path: str) -> None:
        """Actual implementation of loading Datumaro format."""
        self._infos = self._load_infos(self._parsed_anns)
        self._categories = self._load_categories(self._parsed_anns)
        self._items = self._load_items(self._parsed_anns)

    @staticmethod
    def _load_infos(parsed):
        return parsed.get("infos", {})

    @staticmethod
    def _load_categories(parsed):
        categories = {}

        parsed_label_cat = parsed["categories"].get(AnnotationType.label.name)
        if parsed_label_cat:
            label_categories = LabelCategories(attributes=parsed_label_cat.get("attributes", []))
            for item in parsed_label_cat["labels"]:
                label_categories.add(
                    item["name"],
                    parent=item["parent"],
                    attributes=item.get("attributes", []),
                )

            for item in parsed_label_cat.get("label_groups", []):
                label_categories.add_label_group(
                    item["name"], labels=item["labels"], group_type=item["group_type"]
                )

            categories[AnnotationType.label] = label_categories

        parsed_mask_cat = parsed["categories"].get(AnnotationType.mask.name)
        if parsed_mask_cat:
            colormap = {}
            for item in parsed_mask_cat["colormap"]:
                colormap[int(item["label_id"])] = (item["r"], item["g"], item["b"])

            mask_categories = MaskCategories(colormap=colormap)
            categories[AnnotationType.mask] = mask_categories

        parsed_points_cat = parsed["categories"].get(AnnotationType.points.name)
        if parsed_points_cat:
            point_categories = PointsCategories()
            for item in parsed_points_cat["items"]:
                point_categories.add(int(item["label_id"]), item["labels"], joints=item["joints"])

            categories[AnnotationType.points] = point_categories

        return categories

    def _load_items(self, parsed):
        items = []
        for item_desc in parsed["items"]:
            item_id = item_desc["id"]

            media = None
            image_info = item_desc.get("image")
            if image_info:
                image_filename = image_info.get("path") or item_id + DatumaroPath.IMAGE_EXT
                image_path = osp.join(self._images_dir, self._subset, image_filename)
                if not osp.isfile(image_path):
                    # backward compatibility
                    old_image_path = osp.join(self._images_dir, image_filename)
                    if osp.isfile(old_image_path):
                        image_path = old_image_path

                media = Image(path=image_path, size=image_info.get("size"))
                self._media_type = Image

            pcd_info = item_desc.get("point_cloud")
            if media and pcd_info:
                raise DatasetImportError("Dataset cannot contain multiple media types")
            if pcd_info:
                pcd_path = pcd_info.get("path")
                point_cloud = osp.join(self._pcd_dir, self._subset, pcd_path)

                related_images = None
                ri_info = item_desc.get("related_images")
                if ri_info:
                    related_images = [
                        Image(size=ri.get("size"), path=ri.get("path")) for ri in ri_info
                    ]

                media = PointCloud(point_cloud, extra_images=related_images)
                self._media_type = PointCloud

            media_desc = item_desc.get("media")
            if not media and media_desc and media_desc.get("path"):
                media = MediaElement(path=media_desc.get("path"))
                self._media_type = MediaElement

            annotations = self._load_annotations(item_desc)

            item = DatasetItem(
                id=item_id,
                subset=self._subset,
                annotations=annotations,
                media=media,
                attributes=item_desc.get("attr"),
            )

            items.append(item)

        return items

    @staticmethod
    def _load_annotations(item):
        parsed = item["annotations"]
        loaded = []

        for ann in parsed:
            ann_id = ann.get("id")
            ann_type = AnnotationType[ann["type"]]
            attributes = ann.get("attributes")
            group = ann.get("group")

            label_id = ann.get("label_id")
            z_order = ann.get("z_order")
            points = ann.get("points")

            if ann_type == AnnotationType.label:
                loaded.append(Label(label=label_id, id=ann_id, attributes=attributes, group=group))

            elif ann_type == AnnotationType.mask:
                rle = ann["rle"]
                rle["counts"] = rle["counts"].encode("ascii")
                loaded.append(
                    RleMask(
                        rle=rle,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                        z_order=z_order,
                    )
                )

            elif ann_type == AnnotationType.polyline:
                loaded.append(
                    PolyLine(
                        points,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                        z_order=z_order,
                    )
                )

            elif ann_type == AnnotationType.polygon:
                loaded.append(
                    Polygon(
                        points,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                        z_order=z_order,
                    )
                )

            elif ann_type == AnnotationType.bbox:
                x, y, w, h = ann["bbox"]
                loaded.append(
                    Bbox(
                        x,
                        y,
                        w,
                        h,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                        z_order=z_order,
                    )
                )

            elif ann_type == AnnotationType.points:
                loaded.append(
                    Points(
                        points,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                        z_order=z_order,
                    )
                )

            elif ann_type == AnnotationType.caption:
                caption = ann.get("caption")
                loaded.append(Caption(caption, id=ann_id, attributes=attributes, group=group))

            elif ann_type == AnnotationType.cuboid_3d:
                loaded.append(
                    Cuboid3d(
                        ann.get("position"),
                        ann.get("rotation"),
                        ann.get("scale"),
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                    )
                )

            elif ann_type == AnnotationType.ellipse:
                loaded.append(
                    Ellipse(
                        *points,
                        label=label_id,
                        id=ann_id,
                        attributes=attributes,
                        group=group,
                        z_order=z_order,
                    )
                )

            else:
                raise NotImplementedError()

        return loaded



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import logging as log
import os.path as osp
import random
import re
from collections import Counter
from copy import deepcopy
from enum import Enum, auto
from itertools import chain
from typing import Dict, Iterable, List, Optional, Tuple, Union

import cv2
import numpy as np
import pycocotools.mask as mask_utils

import datumaro.util.mask_tools as mask_tools
from datumaro.components.annotation import (
    AnnotationType,
    Bbox,
    Caption,
    Ellipse,
    Label,
    LabelCategories,
    Mask,
    MaskCategories,
    Points,
    PointsCategories,
    Polygon,
    PolyLine,
    RleMask,
)
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME, DatasetInfo, DatasetItem, IDataset
from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image
from datumaro.components.transformer import ItemTransform, Transform
from datumaro.util import NOTSET, filter_dict, parse_str_enum_value, take_by
from datumaro.util.annotation_util import find_group_leader, find_instances


class CropCoveredSegments(ItemTransform, CliPlugin):
    """
    Sorts polygons and masks ("segments") according to `z_order`,
    crops covered areas of underlying segments. If a segment is split
    into several independent parts by the segments above, produces
    the corresponding number of separate annotations joined into a group.
    """

    def transform_item(self, item):
        annotations = []
        segments = []
        for ann in item.annotations:
            if ann.type in {AnnotationType.polygon, AnnotationType.mask}:
                segments.append(ann)
            else:
                annotations.append(ann)
        if not segments:
            return item

        if not isinstance(item.media, Image):
            raise Exception("Image info is required for this transform")
        h, w = item.media.size
        segments = self.crop_segments(segments, w, h)

        annotations += segments
        return self.wrap_item(item, annotations=annotations)

    @classmethod
    def crop_segments(cls, segment_anns, img_width, img_height):
        segment_anns = sorted(segment_anns, key=lambda x: x.z_order)

        segments = []
        for s in segment_anns:
            if s.type == AnnotationType.polygon:
                segments.append(s.points)
            elif s.type == AnnotationType.mask:
                if isinstance(s, RleMask):
                    rle = s.rle
                else:
                    rle = mask_tools.mask_to_rle(s.image)
                segments.append(rle)

        segments = mask_tools.crop_covered_segments(segments, img_width, img_height)

        new_anns = []
        for ann, new_segment in zip(segment_anns, segments):
            fields = {
                "z_order": ann.z_order,
                "label": ann.label,
                "id": ann.id,
                "group": ann.group,
                "attributes": ann.attributes,
            }
            if ann.type == AnnotationType.polygon:
                if fields["group"] is None:
                    fields["group"] = cls._make_group_id(segment_anns + new_anns, fields["id"])
                for polygon in new_segment:
                    new_anns.append(Polygon(points=polygon, **fields))
            else:
                rle = mask_tools.mask_to_rle(new_segment)
                rle = mask_utils.frPyObjects(rle, *rle["size"])
                new_anns.append(RleMask(rle=rle, **fields))

        return new_anns

    @staticmethod
    def _make_group_id(anns, ann_id):
        if ann_id:
            return ann_id
        max_gid = max(anns, default=0, key=lambda x: x.group)
        return max_gid + 1


class MergeInstanceSegments(ItemTransform, CliPlugin):
    """
    Replaces instance masks and, optionally, polygons with a single mask.
    A group of annotations with the same group id is considered an "instance".
    The largest annotation in the group is considered the group "head", so the
    resulting mask takes properties from that annotation.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("--include-polygons", action="store_true", help="Include polygons")
        return parser

    def __init__(self, extractor, include_polygons=False):
        super().__init__(extractor)

        self._include_polygons = include_polygons

    def transform_item(self, item):
        annotations = []
        segments = []
        for ann in item.annotations:
            if ann.type in {AnnotationType.polygon, AnnotationType.mask}:
                segments.append(ann)
            else:
                annotations.append(ann)
        if not segments:
            return item

        if not isinstance(item.media, Image):
            raise Exception("Image info is required for this transform")
        h, w = item.media.size
        instances = self.find_instances(segments)
        segments = [self.merge_segments(i, w, h, self._include_polygons) for i in instances]
        segments = sum(segments, [])

        annotations += segments
        return self.wrap_item(item, annotations=annotations)

    @classmethod
    def merge_segments(cls, instance, img_width, img_height, include_polygons=False):
        polygons = [a for a in instance if a.type == AnnotationType.polygon]
        masks = [a for a in instance if a.type == AnnotationType.mask]
        if not polygons and not masks:
            return []
        if not polygons and len(masks) == 1:
            return masks

        leader = find_group_leader(polygons + masks)
        instance = []

        # Build the resulting mask
        mask = None

        if include_polygons and polygons:
            polygons = [p.points for p in polygons]
            mask = mask_tools.rles_to_mask(polygons, img_width, img_height)
        else:
            instance += polygons  # keep unused polygons

        if masks:
            masks = (m.image for m in masks)
            if mask is not None:
                masks = chain(masks, [mask])
            mask = mask_tools.merge_masks(masks)

        if mask is None:
            return instance

        mask = mask_tools.mask_to_rle(mask)
        mask = mask_utils.frPyObjects(mask, *mask["size"])
        instance.append(
            RleMask(
                rle=mask,
                label=leader.label,
                z_order=leader.z_order,
                id=leader.id,
                attributes=leader.attributes,
                group=leader.group,
            )
        )
        return instance

    @staticmethod
    def find_instances(annotations):
        return find_instances(
            a for a in annotations if a.type in {AnnotationType.polygon, AnnotationType.mask}
        )


class PolygonsToMasks(ItemTransform, CliPlugin):
    _allowed_types = {AnnotationType.polygon, AnnotationType.ellipse}

    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type in self._allowed_types:
                if not isinstance(item.media, Image):
                    raise Exception("Image info is required for this transform")
                h, w = item.media.size
                annotations.append(self.convert_polygon(ann, h, w))
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_polygon(polygon: Union[Polygon, Ellipse], img_h, img_w):
        rle = mask_utils.frPyObjects([polygon.as_polygon()], img_h, img_w)[0]

        return RleMask(
            rle=rle,
            label=polygon.label,
            z_order=polygon.z_order,
            id=polygon.id,
            attributes=polygon.attributes,
            group=polygon.group,
        )


class BoxesToMasks(ItemTransform, CliPlugin):
    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type == AnnotationType.bbox:
                if not isinstance(item.media, Image):
                    raise Exception("Image info is required for this transform")
                h, w = item.media.size
                annotations.append(self.convert_bbox(ann, h, w))
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_bbox(bbox, img_h, img_w):
        rle = mask_utils.frPyObjects([bbox.as_polygon()], img_h, img_w)[0]

        return RleMask(
            rle=rle,
            label=bbox.label,
            z_order=bbox.z_order,
            id=bbox.id,
            attributes=bbox.attributes,
            group=bbox.group,
        )


class MasksToPolygons(ItemTransform, CliPlugin):
    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type == AnnotationType.mask:
                polygons = self.convert_mask(ann)
                if not polygons:
                    log.debug(
                        "[%s]: item %s: "
                        "Mask conversion to polygons resulted in too "
                        "small polygons, which were discarded" % (self.NAME, item.id)
                    )
                annotations.extend(polygons)
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_mask(mask):
        polygons = mask_tools.mask_to_polygons(mask.image)

        return [
            Polygon(
                points=p,
                label=mask.label,
                z_order=mask.z_order,
                id=mask.id,
                attributes=mask.attributes,
                group=mask.group,
            )
            for p in polygons
        ]


class ShapesToBoxes(ItemTransform, CliPlugin):
    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if ann.type in {
                AnnotationType.mask,
                AnnotationType.polygon,
                AnnotationType.polyline,
                AnnotationType.points,
            }:
                annotations.append(self.convert_shape(ann))
            else:
                annotations.append(ann)

        return self.wrap_item(item, annotations=annotations)

    @staticmethod
    def convert_shape(shape):
        bbox = shape.get_bbox()
        return Bbox(
            *bbox,
            label=shape.label,
            z_order=shape.z_order,
            id=shape.id,
            attributes=shape.attributes,
            group=shape.group,
        )


class Reindex(Transform, CliPlugin):
    """
    Replaces dataset item IDs with sequential indices.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("-s", "--start", type=int, default=1, help="Start value for item ids")
        return parser

    def __init__(self, extractor, start=1):
        super().__init__(extractor)
        self._length = "parent"
        self._start = start

    def __iter__(self):
        for i, item in enumerate(self._extractor):
            yield self.wrap_item(item, id=i + self._start)


class MapSubsets(ItemTransform, CliPlugin):
    """
    Renames subsets in the dataset.
    """

    @staticmethod
    def _mapping_arg(s):
        parts = s.split(":")
        if len(parts) != 2:
            raise argparse.ArgumentTypeError()
        return parts

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-s",
            "--subset",
            action="append",
            type=cls._mapping_arg,
            dest="mapping",
            help="Subset mapping of the form: 'src:dst' (repeatable)",
        )
        return parser

    def __init__(self, extractor, mapping=None):
        super().__init__(extractor)

        if mapping is None:
            mapping = {}
        elif not isinstance(mapping, dict):
            mapping = dict(tuple(m) for m in mapping)
        self._mapping = mapping

        if extractor.subsets():
            counts = Counter(mapping.get(s, s) or DEFAULT_SUBSET_NAME for s in extractor.subsets())
            if all(c == 1 for c in counts.values()):
                self._length = "parent"
            self._subsets = set(counts)

    def transform_item(self, item):
        return self.wrap_item(item, subset=self._mapping.get(item.subset, item.subset))


class RandomSplit(Transform, CliPlugin):
    """
    Joins all subsets into one and splits the result into few parts.
    It is expected that item ids are unique and subset ratios sum up to 1.|n
    |n
    Example:|n

    .. code-block::

    |s|s|s|s%(prog)s --subset train:.67 --subset test:.33
    """

    # avoid https://bugs.python.org/issue16399
    _default_split = [("train", 0.67), ("test", 0.33)]

    @staticmethod
    def _split_arg(s):
        parts = s.split(":")
        if len(parts) != 2:
            raise argparse.ArgumentTypeError()
        return (parts[0], float(parts[1]))

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-s",
            "--subset",
            action="append",
            type=cls._split_arg,
            dest="splits",
            help="Subsets in the form: '<subset>:<ratio>' "
            "(repeatable, default: %s)" % dict(cls._default_split),
        )
        parser.add_argument("--seed", type=int, help="Random seed")
        return parser

    def __init__(self, extractor, splits, seed=None):
        super().__init__(extractor)

        if splits is None:
            splits = self._default_split

        assert 0 < len(splits), "Expected at least one split"
        assert all(0.0 <= r and r <= 1.0 for _, r in splits), (
            "Ratios are expected to be in the range [0; 1], but got %s" % splits
        )

        total_ratio = sum(s[1] for s in splits)
        if not abs(total_ratio - 1.0) <= 1e-7:
            raise Exception(
                "Sum of ratios is expected to be 1, got %s, which is %s" % (splits, total_ratio)
            )

        dataset_size = len(extractor)
        indices = list(range(dataset_size))
        random.seed(seed)
        random.shuffle(indices)
        parts = []
        s = 0
        lower_boundary = 0
        for split_idx, (subset, ratio) in enumerate(splits):
            s += ratio
            upper_boundary = int(s * dataset_size)
            if split_idx == len(splits) - 1:
                upper_boundary = dataset_size
            subset_indices = set(indices[lower_boundary:upper_boundary])
            parts.append((subset_indices, subset))
            lower_boundary = upper_boundary
        self._parts = parts

        self._subsets = set(s[0] for s in splits)
        self._length = "parent"

    def _find_split(self, index):
        for subset_indices, subset in self._parts:
            if index in subset_indices:
                return subset
        return subset  # all the possible remainder goes to the last split

    def __iter__(self):
        for i, item in enumerate(self._extractor):
            yield self.wrap_item(item, subset=self._find_split(i))


class IdFromImageName(ItemTransform, CliPlugin):
    """
    Renames items in the dataset using image file name (without extension).
    """

    def transform_item(self, item):
        if isinstance(item.media, Image) and item.media.path:
            name = osp.splitext(osp.basename(item.media.path))[0]
            return self.wrap_item(item, id=name)
        else:
            log.debug("Can't change item id for item '%s': " "item has no image info" % item.id)
            return item


class Rename(ItemTransform, CliPlugin):
    r"""
    Renames items in the dataset. Supports regular expressions.
    The first character in the expression is a delimiter for
    the pattern and replacement parts. Replacement part can also
    contain `str.format` replacement fields with the `item`
    (of type `DatasetItem`) object available.|n
    |n
    Examples:|n
    |s|s- Replace 'pattern' with 'replacement':|n

      .. code-block::

    |s|s|s|srename -e '|pattern|replacement|'|n
    |n
    |s|s- Remove 'frame_' from item ids:|n

      .. code-block::

    |s|s|s|srename -e '|^frame_||'|n
    |n
    |s|s- Rename by regex:|n

      .. code-block::

    |s|s|s|srename -e '|frame_(\d+)_extra|{item.subset}_id_\1|'
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-e",
            "--regex",
            help="Regex for renaming in the form " "'<sep><search><sep><replacement><sep>'",
        )
        return parser

    def __init__(self, extractor, regex):
        super().__init__(extractor)

        assert regex and isinstance(regex, str)
        parts = regex.split(regex[0], maxsplit=3)
        regex, sub = parts[1:3]
        self._re = re.compile(regex)
        self._sub = sub

    def transform_item(self, item):
        return self.wrap_item(item, id=self._re.sub(self._sub, item.id).format(item=item))


class RemapLabels(ItemTransform, CliPlugin):
    """
    Changes labels in the dataset.|n
    |n
    A label can be:|n
    |s|s- renamed (and joined with existing) -|n
    |s|s|s|swhen '--label <old_name>:<new_name>' is specified|n
    |s|s- deleted - when '--label <name>:' is specified, or default action |n
    |s|s|s|sis 'delete' and the label is not mentioned in the list. |n
    |s|s|s|sWhen a label is deleted, all the associated annotations are removed|n
    |s|s- kept unchanged - when specified '--label <name>:<name>'|n
    |s|s|s|sor default action is 'keep' and the label is not mentioned in the list.|n
    |n
    Annotations with no label are managed by the default action policy.|n
    |n
    Examples:|n
    |n
    |s|s- Remove the 'person' label (and corresponding annotations):|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person: --default keep|n
    |n
    |s|s- Rename 'person' to 'pedestrian' and 'human' to 'pedestrian', join:|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person:pedestrian -l human:pedestrian --default keep|n
    |n
    |s|s- Rename 'person' to 'car' and 'cat' to 'dog', keep 'bus', remove others:|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person:car -l bus:bus -l cat:dog --default delete
    """

    class DefaultAction(Enum):
        keep = auto()
        delete = auto()

    @staticmethod
    def _split_arg(s):
        parts = s.split(":")
        if len(parts) != 2:
            raise argparse.ArgumentTypeError()
        return (parts[0], parts[1])

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-l",
            "--label",
            action="append",
            type=cls._split_arg,
            dest="mapping",
            help="Label in the form of: '<src>:<dst>' (repeatable)",
        )
        parser.add_argument(
            "--default",
            choices=[a.name for a in cls.DefaultAction],
            default=cls.DefaultAction.keep.name,
            help="Action for unspecified labels (default: %(default)s)",
        )
        return parser

    def __init__(
        self,
        extractor: IDataset,
        mapping: Union[Dict[str, str], List[Tuple[str, str]]],
        default: Union[None, str, DefaultAction] = None,
    ):
        super().__init__(extractor)

        default = parse_str_enum_value(default, self.DefaultAction, self.DefaultAction.keep)
        self._default_action = default

        assert isinstance(mapping, (dict, list))
        if isinstance(mapping, list):
            mapping = dict(mapping)

        self._categories = {}

        src_categories = self._extractor.categories()

        src_label_cat = src_categories.get(AnnotationType.label)
        if src_label_cat is not None:
            self._make_label_id_map(src_label_cat, mapping, default)

        src_mask_cat = src_categories.get(AnnotationType.mask)
        if src_mask_cat is not None:
            assert src_label_cat is not None
            dst_mask_cat = MaskCategories(attributes=deepcopy(src_mask_cat.attributes))
            for old_id, old_color in src_mask_cat.colormap.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_mask_cat:
                    dst_mask_cat.colormap[new_id] = deepcopy(old_color)

            self._categories[AnnotationType.mask] = dst_mask_cat

        src_point_cat = src_categories.get(AnnotationType.points)
        if src_point_cat is not None:
            assert src_label_cat is not None
            dst_point_cat = PointsCategories(attributes=deepcopy(src_point_cat.attributes))
            for old_id, old_cat in src_point_cat.items.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_point_cat:
                    dst_point_cat.items[new_id] = deepcopy(old_cat)

            self._categories[AnnotationType.points] = dst_point_cat

        assert len(self._categories) == len(src_categories)

    def _make_label_id_map(self, src_label_cat, label_mapping, default_action):
        dst_label_cat = LabelCategories(attributes=deepcopy(src_label_cat.attributes))

        id_mapping = {}
        for src_index, src_label in enumerate(src_label_cat.items):
            dst_label = label_mapping.get(src_label.name, NOTSET)
            if dst_label is NOTSET and default_action == self.DefaultAction.keep:
                dst_label = src_label.name  # keep unspecified as is
            elif not dst_label or dst_label is NOTSET:
                continue

            dst_index = dst_label_cat.find(dst_label)[0]
            if dst_index is None:
                dst_index = dst_label_cat.add(
                    dst_label, src_label.parent, deepcopy(src_label.attributes)
                )
            id_mapping[src_index] = dst_index

        if log.getLogger().isEnabledFor(log.DEBUG):
            log.debug("Label mapping:")
            for src_id, src_label in enumerate(src_label_cat.items):
                if id_mapping.get(src_id) is not None:
                    log.debug(
                        "#%s '%s' -> #%s '%s'",
                        src_id,
                        src_label.name,
                        id_mapping[src_id],
                        dst_label_cat.items[id_mapping[src_id]].name,
                    )
                else:
                    log.debug("#%s '%s' -> <deleted>", src_id, src_label.name)

        self._map_id = lambda src_id: id_mapping.get(src_id, None)

        for label in dst_label_cat:
            if label.parent not in dst_label_cat:
                label.parent = ""
        self._categories[AnnotationType.label] = dst_label_cat

    def categories(self):
        return self._categories

    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if getattr(ann, "label", None) is not None:
                conv_label = self._map_id(ann.label)
                if conv_label is not None:
                    annotations.append(ann.wrap(label=conv_label))
            elif self._default_action is self.DefaultAction.keep:
                annotations.append(ann.wrap())
        return item.wrap(annotations=annotations)


class ProjectInfos(Transform, CliPlugin):
    """
    Changes the content of infos.
    A user can add meta-data of dataset such as author, comments, or related papers.
    Infos values are not affect on the dataset structure.
    We thus can add any meta-data freely.
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-i",
            "--infos",
            action="append",
            dest="dst_infos",
            help="A dictionary of the dataset meta-information",
        )
        parser.add_argument(
            "-o",
            "--overwrite",
            action="store_true",
            dest="overwrite",
            help="Overwrite the infos of src if True",
        )
        return parser

    def __init__(self, extractor: IDataset, dst_infos: DatasetInfo, overwrite: bool = False):
        super().__init__(extractor)

        if overwrite:
            self._infos = dst_infos
        else:
            self._infos = deepcopy(extractor.infos())
            for k, v in dst_infos.items():
                self._infos[k] = v

    def __iter__(self):
        for item in self._extractor:
            if item is not None:
                yield item

    def infos(self):
        return self._infos


class ProjectLabels(ItemTransform):
    """
    Changes the order of labels in the dataset from the existing
    to the desired one, removes unknown labels and adds new labels.
    Updates or removes the corresponding annotations.|n
    |n
    Labels are matched by names (case dependent). Parent labels are only kept
    if they are present in the resulting set of labels. If new labels are
    added, and the dataset has mask colors defined, new labels will obtain
    generated colors.|n
    |n
    Useful for merging similar datasets, whose labels need to be aligned.|n
    |n
    Examples:|n
    |s|s- Align the source dataset labels to [person, cat, dog]:|n

    |s|s.. code-block::

    |s|s|s|s%(prog)s -l person -l cat -l dog
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-l",
            "--label",
            action="append",
            dest="dst_labels",
            help="Label name (repeatable, ordered)",
        )
        return parser

    def __init__(self, extractor: IDataset, dst_labels: Union[Iterable[str], LabelCategories]):
        super().__init__(extractor)

        self._categories = {}

        src_categories = self._extractor.categories()

        src_label_cat = src_categories.get(AnnotationType.label)

        if isinstance(dst_labels, LabelCategories):
            dst_label_cat = deepcopy(dst_labels)
        else:
            dst_labels = list(dst_labels)

            if src_label_cat:
                dst_label_cat = LabelCategories(attributes=deepcopy(src_label_cat.attributes))

                for dst_label in dst_labels:
                    assert isinstance(dst_label, str)
                    src_label = src_label_cat.find(dst_label)[1]
                    if src_label is not None:
                        dst_label_cat.add(
                            dst_label, src_label.parent, deepcopy(src_label.attributes)
                        )
                    else:
                        dst_label_cat.add(dst_label)
            else:
                dst_label_cat = LabelCategories.from_iterable(dst_labels)

        for label in dst_label_cat:
            if label.parent not in dst_label_cat:
                label.parent = ""
        self._categories[AnnotationType.label] = dst_label_cat

        self._make_label_id_map(src_label_cat, dst_label_cat)

        src_mask_cat = src_categories.get(AnnotationType.mask)
        if src_mask_cat is not None:
            assert src_label_cat is not None
            dst_mask_cat = MaskCategories(attributes=deepcopy(src_mask_cat.attributes))
            for old_id, old_color in src_mask_cat.colormap.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_mask_cat:
                    dst_mask_cat.colormap[new_id] = deepcopy(old_color)

            # Generate new colors for new labels, keep old untouched
            existing_colors = set(dst_mask_cat.colormap.values())
            color_bank = iter(
                mask_tools.generate_colormap(len(dst_label_cat), include_background=False).values()
            )
            for new_id, new_label in enumerate(dst_label_cat):
                if new_label.name in src_label_cat:
                    continue
                if new_id in dst_mask_cat:
                    continue

                color = next(color_bank)
                while color in existing_colors:
                    color = next(color_bank)

                dst_mask_cat.colormap[new_id] = color

            self._categories[AnnotationType.mask] = dst_mask_cat

        src_point_cat = src_categories.get(AnnotationType.points)
        if src_point_cat is not None:
            assert src_label_cat is not None
            dst_point_cat = PointsCategories(attributes=deepcopy(src_point_cat.attributes))
            for old_id, old_cat in src_point_cat.items.items():
                new_id = self._map_id(old_id)
                if new_id is not None and new_id not in dst_point_cat:
                    dst_point_cat.items[new_id] = deepcopy(old_cat)

            self._categories[AnnotationType.points] = dst_point_cat

    def _make_label_id_map(self, src_label_cat, dst_label_cat):
        id_mapping = {
            src_id: dst_label_cat.find(src_label_cat[src_id].name)[0]
            for src_id in range(len(src_label_cat or ()))
        }
        self._map_id = lambda src_id: id_mapping.get(src_id, None)

    def categories(self):
        return self._categories

    def transform_item(self, item):
        annotations = []
        for ann in item.annotations:
            if getattr(ann, "label", None) is not None:
                conv_label = self._map_id(ann.label)
                if conv_label is not None:
                    annotations.append(ann.wrap(label=conv_label))
            else:
                annotations.append(ann.wrap())
        return item.wrap(annotations=annotations)


class AnnsToLabels(ItemTransform, CliPlugin):
    """
    Collects all labels from annotations (of all types) and
    transforms them into a set of annotations of type Label
    """

    def transform_item(self, item):
        labels = set(p.label for p in item.annotations if getattr(p, "label") is not None)
        annotations = []
        for label in labels:
            annotations.append(Label(label=label))

        return item.wrap(annotations=annotations)


class BboxValuesDecrement(ItemTransform, CliPlugin):
    """
    Subtracts one from the coordinates of bounding boxes
    """

    def transform_item(self, item):
        annotations = [p for p in item.annotations if p.type != AnnotationType.bbox]
        bboxes = [p for p in item.annotations if p.type == AnnotationType.bbox]
        for bbox in bboxes:
            annotations.append(
                Bbox(
                    bbox.x - 1,
                    bbox.y - 1,
                    bbox.w,
                    bbox.h,
                    label=bbox.label,
                    attributes=bbox.attributes,
                )
            )

        return item.wrap(annotations=annotations)


class ResizeTransform(ItemTransform):
    """
    Resizes images and annotations in the dataset to the specified size.
    Supports upscaling, downscaling and mixed variants.|n
    |n
    Examples:|n
        - Resize all images to 256x256 size|n

        .. code-block::

        |s|s%(prog)s -dw 256 -dh 256
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("-dw", "--width", type=int, help="Destination image width")
        parser.add_argument("-dh", "--height", type=int, help="Destination image height")
        return parser

    def __init__(self, extractor: IDataset, width: int, height: int) -> None:
        super().__init__(extractor)

        assert width > 0 and height > 0
        self._width = width
        self._height = height

    @staticmethod
    def _lazy_resize_image(image, new_size):
        def _resize_image(_):
            h, w = image.size
            yscale = new_size[0] / float(h)
            xscale = new_size[1] / float(w)

            # LANCZOS4 is preferable for upscaling, but it works quite slow
            method = cv2.INTER_AREA if (xscale * yscale) < 1 else cv2.INTER_CUBIC

            resized_image = cv2.resize(image.data / 255.0, new_size[::-1], interpolation=method)
            resized_image *= 255.0
            return resized_image

        return Image(_resize_image, ext=image.ext, size=new_size)

    @staticmethod
    def _lazy_resize_mask(mask, new_size):
        def _resize_image():
            # Can use only NEAREST for masks,
            # because we can't have interpolated values
            rescaled_mask = cv2.resize(
                mask.image.astype(np.float32), new_size[::-1], interpolation=cv2.INTER_NEAREST
            )
            return rescaled_mask.astype(np.uint8)

        return _resize_image

    def transform_item(self, item):
        if not isinstance(item.media, Image):
            raise DatumaroError(
                "Item %s: image info is required for this " "transform" % (item.id,)
            )

        h, w = item.media.size
        xscale = self._width / float(w)
        yscale = self._height / float(h)

        new_size = (self._height, self._width)

        resized_image = None
        if item.media.has_data:
            resized_image = self._lazy_resize_image(item.media, new_size)

        resized_annotations = []
        for ann in item.annotations:
            if isinstance(ann, Bbox):
                resized_annotations.append(
                    ann.wrap(
                        x=ann.x * xscale,
                        y=ann.y * yscale,
                        w=ann.w * xscale,
                        h=ann.h * yscale,
                    )
                )
            elif isinstance(ann, (Polygon, Points, PolyLine)):
                resized_annotations.append(
                    ann.wrap(
                        points=[
                            p
                            for t in ((x * xscale, y * yscale) for x, y in take_by(ann.points, 2))
                            for p in t
                        ]
                    )
                )
            elif isinstance(ann, Mask):
                rescaled_mask = self._lazy_resize_mask(ann, new_size)
                resized_annotations.append(ann.wrap(image=rescaled_mask))
            elif isinstance(ann, (Caption, Label)):
                resized_annotations.append(ann)
            else:
                assert False, f"Unexpected annotation type {type(ann)}"

        return self.wrap_item(item, media=resized_image, annotations=resized_annotations)


class RemoveItems(ItemTransform):
    """
    Allows to remove specific dataset items from dataset by their ids.|n
    |n
    Can be useful to clean the dataset from broken or unnecessary samples.|n
    |n
    Examples:|n
        - Remove specific items from the dataset|n

        .. code-block::

        |s|s%(prog)s --id 'image1:train' --id 'image2:test'
    """

    @staticmethod
    def _parse_id(s):
        full_id = s.split(":")
        if len(full_id) != 2:
            raise argparse.ArgumentTypeError(
                None, message="Invalid id format of '%s'. " "Expected a 'name:subset' pair." % s
            )
        return full_id

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--id",
            dest="ids",
            type=cls._parse_id,
            action="append",
            required=True,
            help="Item id to remove. Id is 'name:subset' pair (repeatable)",
        )
        return parser

    def __init__(self, extractor: IDataset, ids: Iterable[Tuple[str, str]]):
        super().__init__(extractor)
        self._ids = set(tuple(v) for v in (ids or []))

    def transform_item(self, item):
        if (item.id, item.subset) in self._ids:
            return None
        return item


class RemoveAnnotations(ItemTransform):
    """
    Allows to remove annotations on specific dataset items.|n
    |n
    Can be useful to clean the dataset from broken or unnecessary annotations.|n
    |n
    Examples:|n
        - Remove annotations from specific items in the dataset|n

        .. code-block::

        |s|s%(prog)s --id 'image1:train' --id 'image2:test'
    """

    @staticmethod
    def _parse_id(s):
        full_id = s.split(":")
        if len(full_id) != 2:
            raise argparse.ArgumentTypeError(
                None, message="Invalid id format of '%s'. " "Expected a 'name:subset' pair." % s
            )
        return full_id

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--id",
            dest="ids",
            type=cls._parse_id,
            action="append",
            help="Image id to clean from annotations. "
            "Id is 'name:subset' pair. If not specified, removes "
            "all annotations (repeatable)",
        )
        return parser

    def __init__(self, extractor: IDataset, *, ids: Optional[Iterable[Tuple[str, str]]] = None):
        super().__init__(extractor)
        self._ids = set(tuple(v) for v in (ids or []))

    def transform_item(self, item: DatasetItem):
        if not self._ids or (item.id, item.subset) in self._ids:
            return item.wrap(annotations=[])
        return item


class RemoveAttributes(ItemTransform):
    """
    Allows to remove item and annotation attributes in a dataset.|n
    |n
    Can be useful to clean the dataset from broken or unnecessary attributes.|n
    |n
    Examples:|n
        - Remove the `is_crowd` attribute from dataset|n

        .. code-block::

        |s|s%(prog)s --attr 'is_crowd'|n
        |n
        - Remove the `occluded` attribute from annotations of|n
        |s|sthe `2010_001705` item in the `train` subset|n

        .. code-block::

        |s|s%(prog)s --id '2010_001705:train' --attr 'occluded'
    """

    @staticmethod
    def _parse_id(s):
        full_id = s.split(":")
        if len(full_id) != 2:
            raise argparse.ArgumentTypeError(
                None, message="Invalid id format of '%s'. " "Expected a 'name:subset' pair." % s
            )
        return full_id

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--id",
            dest="ids",
            type=cls._parse_id,
            action="append",
            help="Image id to clean from annotations. "
            "Id is 'name:subset' pair. If not specified, "
            "affects all images and annotations (repeatable)",
        )
        parser.add_argument(
            "-a",
            "--attr",
            action="append",
            dest="attributes",
            help="Attribute name to be removed. If not specified, "
            "removes all attributes (repeatable)",
        )
        return parser

    def __init__(
        self,
        extractor: IDataset,
        ids: Optional[Iterable[Tuple[str, str]]] = None,
        attributes: Optional[Iterable[str]] = None,
    ):
        super().__init__(extractor)
        self._ids = set(tuple(v) for v in (ids or []))
        self._attributes = set(attributes or [])

    def _filter_attrs(self, attrs):
        if not self._attributes:
            return None
        else:
            return filter_dict(attrs, exclude_keys=self._attributes)

    def transform_item(self, item: DatasetItem):
        if not self._ids or (item.id, item.subset) in self._ids:
            filtered_annotations = []
            for ann in item.annotations:
                filtered_annotations.append(ann.wrap(attributes=self._filter_attrs(ann.attributes)))

            return item.wrap(
                attributes=self._filter_attrs(item.attributes), annotations=filtered_annotations
            )
        return item



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
import struct
from typing import Dict, List, Optional, Tuple

import numpy as np
import pycocotools.mask as mask_utils

from datumaro.components.annotation import (
    Annotation,
    AnnotationType,
    Bbox,
    Caption,
    Cuboid3d,
    Ellipse,
    Label,
    Mask,
    Points,
    Polygon,
    PolyLine,
    RleMask,
    _Shape,
)

from .common import DictMapper, FloatListMapper, IntListMapper, Mapper, StringMapper

MAGIC_NUM_FOR_NONE = 2**31 - 1


class AnnotationMapper(Mapper):
    ann_type = AnnotationType.unknown

    @classmethod
    def forward(cls, ann: Annotation) -> bytes:
        _bytearray = bytearray()
        _bytearray.extend(struct.pack("<Bqq", ann.type, ann.id, ann.group))
        _bytearray.extend(DictMapper.forward(ann.attributes))
        return bytes(_bytearray)

    @classmethod
    def backward_dict(cls, _bytes: bytes, offset: int = 0) -> Tuple[Dict, int]:
        _, id, group = struct.unpack_from("<Bqq", _bytes, offset)
        offset += 17  # struct.calcsize("<Bqq") = 17
        attributes, offset = DictMapper.backward(_bytes, offset)
        return {"id": id, "attributes": attributes, "group": group}, offset

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Annotation, int]:
        ann_dict, offset = cls.backward_dict(_bytes, offset)
        return Annotation(**ann_dict), offset

    @staticmethod
    def forward_optional_label(label: Optional[int]) -> int:
        return label if label is not None else MAGIC_NUM_FOR_NONE

    @staticmethod
    def backward_optional_label(label: int) -> Optional[int]:
        return label if label != MAGIC_NUM_FOR_NONE else None

    @staticmethod
    def parse_ann_type(_bytes: bytes, offset: int = 0) -> AnnotationType:
        (ann_type,) = struct.unpack_from("<B", _bytes, offset)
        return AnnotationType(ann_type)


class LabelMapper(AnnotationMapper):
    ann_type = AnnotationType.label

    @classmethod
    def forward(cls, ann: Label) -> bytes:
        _bytearray = bytearray()
        _bytearray.extend(super().forward(ann))
        _bytearray.extend(struct.pack("<i", ann.label))
        return bytes(_bytearray)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Label, int]:
        ann_dict, offset = super().backward_dict(_bytes, offset)
        (label,) = struct.unpack_from("<i", _bytes, offset)
        offset += 4
        return Label(label=label, **ann_dict), offset


class MaskMapper(AnnotationMapper):
    ann_type = AnnotationType.mask

    @classmethod
    def forward(cls, ann: Mask) -> bytes:
        if isinstance(ann, RleMask):
            rle = ann.rle
        else:
            rle = mask_utils.encode(np.require(ann.image, dtype=np.uint8, requirements="F"))

        h, w = rle["size"]
        counts = rle["counts"]
        len_counts = len(counts)

        _bytearray = bytearray()
        _bytearray.extend(super().forward(ann))
        _bytearray.extend(struct.pack("<ii", cls.forward_optional_label(ann.label), ann.z_order))
        _bytearray.extend(struct.pack(f"<iiI{len_counts}s", h, w, len_counts, counts))
        return bytes(_bytearray)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Mask, int]:
        ann_dict, offset = super().backward_dict(_bytes, offset)
        label, z_order = struct.unpack_from("<ii", _bytes, offset)
        label = cls.backward_optional_label(label)
        offset += 8
        h, w, len_counts = struct.unpack_from("<iiI", _bytes, offset)
        offset += 12
        (counts,) = struct.unpack_from(f"<{len_counts}s", _bytes, offset)
        offset += len_counts

        return (
            RleMask(
                rle={"size": [h, w], "counts": counts}, label=label, z_order=z_order, **ann_dict
            ),
            offset,
        )


class RleMaskMapper(MaskMapper):
    """Just clone MaskMapper."""


class _ShapeMapper(AnnotationMapper):
    @classmethod
    def forward(cls, ann: _Shape) -> bytes:
        _bytearray = bytearray()
        _bytearray.extend(super().forward(ann))
        _bytearray.extend(struct.pack("<ii", cls.forward_optional_label(ann.label), ann.z_order))
        _bytearray.extend(FloatListMapper.forward(ann.points))
        return bytes(_bytearray)

    @classmethod
    def backward_dict(cls, _bytes: bytes, offset: int = 0) -> Tuple[Dict, int]:
        ann_dict, offset = super().backward_dict(_bytes, offset)
        label, z_order = struct.unpack_from("<ii", _bytes, offset)
        offset += 8
        points, offset = FloatListMapper.backward(_bytes, offset)
        return {
            "points": points,
            "label": cls.backward_optional_label(label),
            "z_order": z_order,
            **ann_dict,
        }, offset

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[_Shape, int]:
        ann_dict, offset = cls.backward_dict(_bytes, offset)
        return _Shape(**ann_dict), offset


class PointsMapper(_ShapeMapper):
    ann_type = AnnotationType.points

    @classmethod
    def forward(cls, ann: Points) -> bytes:
        _bytearray = bytearray()
        _bytearray.extend(super().forward(ann))
        _bytearray.extend(IntListMapper.forward(ann.visibility))
        return bytes(_bytearray)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Points, int]:
        shape_dict, offset = super().backward_dict(_bytes, offset)
        visibility, offset = IntListMapper.backward(_bytes, offset)
        return Points(visibility=[Points.Visibility(v) for v in visibility], **shape_dict), offset


class PolyLineMapper(_ShapeMapper):
    ann_type = AnnotationType.polyline

    @classmethod
    def forward(cls, ann: PolyLine) -> bytes:
        return super().forward(ann)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[PolyLine, int]:
        shape_dict, offset = super().backward_dict(_bytes, offset)
        return PolyLine(**shape_dict), offset


class PolygonMapper(_ShapeMapper):
    ann_type = AnnotationType.polygon

    @classmethod
    def forward(cls, ann: Polygon) -> bytes:
        return super().forward(ann)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Polygon, int]:
        shape_dict, offset = super().backward_dict(_bytes, offset)
        return Polygon(**shape_dict), offset


class BboxMapper(_ShapeMapper):
    ann_type = AnnotationType.bbox

    @classmethod
    def forward(cls, ann: Bbox) -> bytes:
        return super().forward(ann)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Bbox, int]:
        shape_dict, offset = super().backward_dict(_bytes, offset)
        x, y, x2, y2 = shape_dict["points"]
        return Bbox(x, y, x2 - x, y2 - y, **shape_dict), offset


class CaptionMapper(AnnotationMapper):
    ann_type = AnnotationType.caption

    @classmethod
    def forward(cls, ann: Caption) -> bytes:
        _bytearray = bytearray()
        _bytearray.extend(super().forward(ann))
        _bytearray.extend(StringMapper.forward(ann.caption))
        return bytes(_bytearray)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Caption, int]:
        ann_dict, offset = super().backward_dict(_bytes, offset)
        caption, offset = StringMapper.backward(_bytes, offset)
        return Caption(caption=caption, **ann_dict), offset


class Cuboid3dMapper(AnnotationMapper):
    ann_type = AnnotationType.cuboid_3d

    @classmethod
    def forward(cls, ann: Cuboid3d) -> bytes:
        _bytearray = bytearray()
        _bytearray.extend(super().forward(ann))
        _bytearray.extend(struct.pack("<i", cls.forward_optional_label(ann.label)))
        _bytearray.extend(FloatListMapper.forward(ann._points))
        return bytes(_bytearray)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Cuboid3d, int]:
        ann_dict, offset = super().backward_dict(_bytes, offset)
        (label,) = struct.unpack_from("<i", _bytes, offset)
        offset += 4
        points, offset = FloatListMapper.backward(_bytes, offset)
        return (
            Cuboid3d(
                position=points[:3],
                rotation=points[3:6],
                scale=points[6:],
                label=cls.backward_optional_label(label),
                **ann_dict,
            ),
            offset,
        )


class EllipseMapper(_ShapeMapper):
    ann_type = AnnotationType.ellipse

    @classmethod
    def forward(cls, ann: Ellipse) -> bytes:
        return super().forward(ann)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[Ellipse, int]:
        shape_dict, offset = super().backward_dict(_bytes, offset)
        x, y, x2, y2 = shape_dict["points"]
        return Ellipse(x, y, x2, y2, **shape_dict), offset


class AnnotationListMapper(Mapper):
    backward_map = {
        AnnotationType.label: LabelMapper.backward,
        AnnotationType.mask: MaskMapper.backward,
        AnnotationType.points: PointsMapper.backward,
        AnnotationType.polyline: PolyLineMapper.backward,
        AnnotationType.polygon: PolygonMapper.backward,
        AnnotationType.bbox: BboxMapper.backward,
        AnnotationType.caption: CaptionMapper.backward,
        AnnotationType.cuboid_3d: Cuboid3dMapper.backward,
        AnnotationType.ellipse: EllipseMapper.backward,
    }

    @classmethod
    def forward(cls, anns: List[Annotation]) -> bytes:
        _bytearray = bytearray()
        _bytearray.extend(struct.pack("<I", len(anns)))

        for ann in anns:
            if isinstance(ann, Label):
                _bytearray.extend(LabelMapper.forward(ann))
            elif isinstance(ann, Mask):
                _bytearray.extend(MaskMapper.forward(ann))
            elif isinstance(ann, Points):
                _bytearray.extend(PointsMapper.forward(ann))
            elif isinstance(ann, PolyLine):
                _bytearray.extend(PolyLineMapper.forward(ann))
            elif isinstance(ann, Polygon):
                _bytearray.extend(PolygonMapper.forward(ann))
            elif isinstance(ann, Bbox):
                _bytearray.extend(BboxMapper.forward(ann))
            elif isinstance(ann, Caption):
                _bytearray.extend(CaptionMapper.forward(ann))
            elif isinstance(ann, Cuboid3d):
                _bytearray.extend(Cuboid3dMapper.forward(ann))
            elif isinstance(ann, Ellipse):
                _bytearray.extend(EllipseMapper.forward(ann))
            else:
                raise NotImplementedError()

        return bytes(_bytearray)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[List[Annotation], int]:
        (n_anns,) = struct.unpack_from("<I", _bytes, offset)
        offset += 4
        anns = []
        for _ in range(n_anns):
            ann_type = AnnotationMapper.parse_ann_type(_bytes, offset)
            ann, offset = cls.backward_map[ann_type](_bytes, offset)
            anns.append(ann)

        return anns, offset



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import glob
import importlib
import logging as log
import os.path as osp
from functools import partial
from inspect import isclass
from typing import Callable, Dict, Generic, Iterable, Iterator, List, Optional, Set, Type, TypeVar

from datumaro.components.cli_plugin import CliPlugin, plugin_types
from datumaro.components.format_detection import (
    DetectedFormat,
    FormatDetectionConfidence,
    RejectionReason,
    detect_dataset_format,
)
from datumaro.util.os_util import import_foreign_module, split_path

T = TypeVar("T")


class Registry(Generic[T]):
    def __init__(self):
        self.items: Dict[str, T] = {}

    def register(self, name: str, value: T) -> T:
        self.items[name] = value
        return value

    def unregister(self, name: str) -> Optional[T]:
        return self.items.pop(name, None)

    def get(self, key: str):
        """Returns a class or a factory function"""
        return self.items[key]

    def __getitem__(self, key: str) -> T:
        return self.get(key)

    def __contains__(self, key) -> bool:
        return key in self.items

    def __iter__(self) -> Iterator[str]:
        return iter(self.items)


class PluginRegistry(Registry[Type[CliPlugin]]):
    def __init__(
        self, filter: Callable[[Type[CliPlugin]], bool] = None
    ):  # pylint: disable=redefined-builtin
        super().__init__()
        self._filter = filter

    def batch_register(self, values: Iterable[CliPlugin]):
        for v in values:
            if self._filter and not self._filter(v):
                continue

            self.register(v.NAME, v)


class Environment:
    _builtin_plugins = None

    @classmethod
    def _make_filter(cls, accept, skip=None):
        accept = (accept,) if isclass(accept) else tuple(accept)
        skip = {skip} if isclass(skip) else set(skip or [])
        skip = tuple(skip | set(accept))
        return partial(cls._check_type, accept=accept, skip=skip)

    @staticmethod
    def _check_type(t, *, accept, skip):
        return issubclass(t, accept) and t not in skip

    def __init__(self):
        from datumaro.components.dataset_base import DatasetBase, SubsetBase
        from datumaro.components.exporter import Exporter
        from datumaro.components.generator import DatasetGenerator
        from datumaro.components.importer import Importer
        from datumaro.components.launcher import Launcher
        from datumaro.components.transformer import ItemTransform, Transform
        from datumaro.components.validator import Validator

        _filter = self._make_filter
        self._extractors = PluginRegistry(_filter(DatasetBase, skip=SubsetBase))
        self._importers = PluginRegistry(_filter(Importer))
        self._launchers = PluginRegistry(_filter(Launcher))
        self._exporters = PluginRegistry(_filter(Exporter))
        self._generators = PluginRegistry(_filter(DatasetGenerator))
        self._transforms = PluginRegistry(_filter(Transform, skip=ItemTransform))
        self._validators = PluginRegistry(_filter(Validator))
        self._builtins_initialized = False

    def _get_plugin_registry(self, name):
        if not self._builtins_initialized:
            self._builtins_initialized = True
            self._register_builtin_plugins()
        return getattr(self, name)

    @property
    def extractors(self) -> PluginRegistry:
        return self._get_plugin_registry("_extractors")

    @property
    def importers(self) -> PluginRegistry:
        return self._get_plugin_registry("_importers")

    @property
    def launchers(self) -> PluginRegistry:
        return self._get_plugin_registry("_launchers")

    @property
    def exporters(self) -> PluginRegistry:
        return self._get_plugin_registry("_exporters")

    @property
    def generators(self) -> PluginRegistry:
        return self._get_plugin_registry("_generators")

    @property
    def transforms(self) -> PluginRegistry:
        return self._get_plugin_registry("_transforms")

    @property
    def validators(self) -> PluginRegistry:
        return self._get_plugin_registry("_validators")

    @staticmethod
    def _find_plugins(plugins_dir):
        plugins = []

        for pattern in ("*.py", "*/*.py", "*/*/*.py"):
            for path in glob.glob(osp.join(glob.escape(plugins_dir), pattern)):
                if not osp.isfile(path):
                    continue

                path_rel = osp.relpath(path, plugins_dir)
                name_parts = split_path(osp.splitext(path_rel)[0])

                # a module with a dot in the name won't load correctly
                if any("." in part for part in name_parts):
                    log.warning(
                        "Python file '%s' in directory '%s' can't be imported "
                        "due to a dot in the name; skipping.",
                        path_rel,
                        plugins_dir,
                    )
                    continue
                plugins.append(".".join(name_parts))

        return plugins

    @classmethod
    def _get_plugin_exports(cls, module, types):
        exports = []
        if hasattr(module, "exports"):
            exports = module.exports
        else:
            for symbol in dir(module):
                if symbol.startswith("_"):
                    continue
                exports.append(getattr(module, symbol))

        exports = [s for s in exports if isclass(s) and issubclass(s, types) and not s in types]

        return exports

    @classmethod
    def _load_plugins(cls, module_names, *, importer, types=None):
        types = tuple(types or plugin_types())

        all_exports = []
        for module_name in module_names:
            try:
                module = importer(module_name)
                exports = cls._get_plugin_exports(module, types)
            except Exception as e:
                module_search_error = ModuleNotFoundError

                message = ["Failed to import module '%s': %s", module_name, e]
                if isinstance(e, module_search_error):
                    log.debug(*message)
                else:
                    log.warning(*message)
                continue

            log.debug(
                "Imported the following symbols from %s: %s"
                % (module_name, ", ".join(s.__name__ for s in exports))
            )
            all_exports.extend(exports)

        return all_exports

    @classmethod
    def _load_builtin_plugins(cls):
        if cls._builtin_plugins is None:
            import datumaro.plugins

            plugins_dir = osp.dirname(datumaro.plugins.__file__)
            module_names = [
                datumaro.plugins.__name__ + "." + name for name in cls._find_plugins(plugins_dir)
            ]
            cls._builtin_plugins = cls._load_plugins(module_names, importer=importlib.import_module)
        return cls._builtin_plugins

    def load_plugins(self, plugins_dir):
        module_names = self._find_plugins(plugins_dir)
        plugins = self._load_plugins(
            module_names, importer=partial(import_foreign_module, path=plugins_dir)
        )

        self._register_plugins(plugins)

    def _register_builtin_plugins(self):
        self._register_plugins(self._load_builtin_plugins())

    def _register_plugins(self, plugins):
        self.extractors.batch_register(plugins)
        self.importers.batch_register(plugins)
        self.launchers.batch_register(plugins)
        self.exporters.batch_register(plugins)
        self.generators.batch_register(plugins)
        self.transforms.batch_register(plugins)
        self.validators.batch_register(plugins)

    def make_extractor(self, name, *args, **kwargs):
        return self.extractors.get(name)(*args, **kwargs)

    def make_importer(self, name, *args, **kwargs):
        return self.importers.get(name)(*args, **kwargs)

    def make_launcher(self, name, *args, **kwargs):
        return self.launchers.get(name)(*args, **kwargs)

    def make_exporter(self, name, *args, **kwargs):
        result = self.exporters.get(name)
        if isclass(result):
            result = result.convert
        return partial(result, *args, **kwargs)

    def make_transform(self, name, *args, **kwargs):
        return partial(self.transforms.get(name), *args, **kwargs)

    def is_format_known(self, name):
        return name in self.importers or name in self.extractors

    def detect_dataset(
        self,
        path: str,
        depth: int = 1,
        rejection_callback: Optional[Callable[[str, RejectionReason, str], None]] = None,
    ) -> List[str]:
        ignore_dirs = {"__MSOSX", "__MACOSX"}
        all_matched_formats: Set[DetectedFormat] = set()

        for _ in range(depth + 1):
            detected_formats = detect_dataset_format(
                (
                    (format_name, importer.detect)
                    for format_name, importer in self.importers.items.items()
                ),
                path,
                rejection_callback=rejection_callback,
            )

            if detected_formats:
                all_matched_formats |= set(detected_formats)

            paths = glob.glob(osp.join(path, "*"))
            path = "" if len(paths) != 1 else paths[0]
            if not osp.isdir(path) or osp.basename(path) in ignore_dirs:
                break

        max_conf = (
            max(all_matched_formats).confidence
            if len(all_matched_formats) > 0
            else FormatDetectionConfidence.NONE
        )

        return [str(format) for format in all_matched_formats if format.confidence == max_conf]

    def __reduce__(self):
        return (self.__class__, ())


DEFAULT_ENVIRONMENT = Environment()



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import struct
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple

from datumaro.util import dump_json, parse_json


class Mapper(ABC):
    @staticmethod
    @abstractmethod
    def forward(obj: Any) -> bytes:
        """Convert an object to bytes."""

    @staticmethod
    @abstractmethod
    def backward(_bytes: bytes, offset: int = 0) -> Tuple[Any, int]:
        """Build an object from bytes."""


class StringMapper(Mapper):
    @staticmethod
    def forward(obj: str) -> bytes:
        obj = obj.encode()
        length = len(obj)
        return struct.pack(f"<I{length}s", length, obj)

    @staticmethod
    def backward(_bytes: bytes, offset: int = 0) -> Tuple[str, int]:
        length = struct.unpack_from("<I", _bytes, offset)[0]
        offset += 4
        string = struct.unpack_from(f"<{length}s", _bytes, offset)[0].decode()
        return string, offset + length


class ListMapper(Mapper):
    _format = ""

    @classmethod
    def forward(cls, obj: List[Any]) -> bytes:
        length = len(obj)
        return struct.pack(f"I{length}{cls._format}", length, *obj)

    @classmethod
    def backward(cls, _bytes: bytes, offset: int = 0) -> Tuple[List[Any], int]:
        (length,) = struct.unpack_from("<I", _bytes, offset)
        offset += 4
        obj = struct.unpack_from(f"<{length}{cls._format}", _bytes, offset)
        offset += 4 * length
        return obj, offset


class IntListMapper(ListMapper):
    _format = "i"


class FloatListMapper(ListMapper):
    _format = "f"


class DictMapper(Mapper):
    @staticmethod
    def forward(obj: Dict[str, Any]) -> bytes:
        if len(obj) == 0:
            msg = b""
        else:
            msg = dump_json(obj)
        length = len(msg)
        return struct.pack(f"<I{length}s", length, msg)

    @staticmethod
    def backward(_bytes: bytes, offset: int = 0) -> Tuple[Dict[str, Any], int]:
        length = struct.unpack_from("<I", _bytes, offset)[0]
        offset += 4
        if length == 0:
            parsed_dict = {}
        else:
            parsed_dict = parse_json(_bytes[offset : offset + length])
        return parsed_dict, offset + length



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from collections import OrderedDict

import numpy as np

from datumaro.components.annotation import AnnotationType, LabelCategories, Mask, MaskCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.components.media import Image
from datumaro.util.image import find_images, load_image
from datumaro.util.mask_tools import generate_colormap, lazy_mask
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class SynthiaPath:
    IMAGES_DIR = "RGB"
    LABELS_SEGM_DIR = "GT/LABELS"
    SEMANTIC_SEGM_DIR = "GT/COLOR"
    LABELMAP_FILE = "label_colors.txt"


SYNTHIA_LABEL_MAP = OrderedDict(
    [
        ("Void", (0, 0, 0)),
        ("Sky", (128, 128, 128)),
        ("Building", (128, 0, 0)),
        ("Road", (128, 64, 128)),
        ("Sidewalk", (0, 0, 192)),
        ("Fence", (64, 64, 128)),
        ("Vegetation", (128, 128, 0)),
        ("Pole", (192, 192, 128)),
        ("Car", (64, 0, 128)),
        ("TrafficSign", (192, 128, 128)),
        ("Pedestrian", (64, 64, 0)),
        ("Bicycle", (0, 128, 192)),
        ("Lanemarking", (0, 172, 0)),
        ("Reserved_1", (0, 0, 0)),
        ("Reserved_2", (0, 0, 0)),
        ("TrafficLight", (0, 128, 128)),
    ]
)


def make_categories(label_map=None):
    if label_map is None:
        label_map = SYNTHIA_LABEL_MAP

    categories = {}
    label_categories = LabelCategories()
    for label in label_map:
        label_categories.add(label)
    categories[AnnotationType.label] = label_categories

    has_colors = any(v is not None for v in label_map.values())
    if not has_colors:  # generate new colors
        colormap = generate_colormap(len(label_map))
    else:  # only copy defined colors
        colormap = {
            label_id: (desc[0], desc[1], desc[2])
            for label_id, desc in enumerate(label_map.values())
        }
    mask_categories = MaskCategories(colormap)
    mask_categories.inverse_colormap  # pylint: disable=pointless-statement
    categories[AnnotationType.mask] = mask_categories
    return categories


def parse_label_map(path):
    label_map = OrderedDict()
    with open(path, "r", encoding="utf-8") as f:
        for line in f:
            # skip empty and commented lines
            line = line.strip()
            if not line or line[0] == "#":
                continue

            # color, name
            label_desc = line.split()

            if 2 < len(label_desc):
                name = label_desc[3]
                color = tuple([int(c) for c in label_desc[:3]])
            else:
                name = label_desc[0]
                color = None

            if name in label_map:
                raise ValueError("Label '%s' is already defined" % name)

            label_map[name] = color
    return label_map


class SynthiaBase(SubsetBase):
    def __init__(self, path):
        if not osp.isdir(path):
            raise FileNotFoundError("Can't read dataset directory '%s'" % path)

        super().__init__()

        self._categories = self._load_categories(path)
        self._items = list(self._load_items(path).values())

    def _load_categories(self, path):
        if has_meta_file(path):
            return make_categories(parse_meta_file(path))
        label_map_path = osp.join(path, SynthiaPath.LABELMAP_FILE)
        if osp.isfile(label_map_path):
            label_map = parse_label_map(label_map_path)
        else:
            label_map = SYNTHIA_LABEL_MAP
        return make_categories(label_map)

    def _load_items(self, root_dir):
        image_dir = osp.join(root_dir, SynthiaPath.IMAGES_DIR)
        if osp.isdir(image_dir):
            images = {
                osp.splitext(osp.relpath(p, image_dir))[0].replace("\\", "/"): p
                for p in find_images(image_dir, recursive=True)
            }
        else:
            images = {}

        items = {}

        inst_dir = osp.join(root_dir, SynthiaPath.LABELS_SEGM_DIR)
        if osp.isdir(inst_dir):
            gt_images = find_images(inst_dir, recursive=True)
            for gt_img in gt_images:
                item_id = osp.splitext(osp.relpath(gt_img, inst_dir))[0].replace("\\", "/")

                anno = []
                labels_mask = load_image(gt_img, dtype=np.uint16)
                dynamic_objects = np.unique(labels_mask[:, :, 1])
                labels_mask = labels_mask[:, :, 2]
                segm_ids = np.unique(labels_mask)
                for segm_id in segm_ids:
                    attr = {"dynamic_object": False}
                    if segm_id != 0 and segm_id in dynamic_objects:
                        attr["dynamic_object"] = True
                    anno.append(
                        Mask(
                            image=self._lazy_extract_mask(labels_mask, segm_id),
                            label=segm_id,
                            attributes=attr,
                        )
                    )

                image = images.get(item_id)
                if image:
                    image = Image(path=image)

                items[item_id] = DatasetItem(id=item_id, media=image, annotations=anno)

        elif osp.isdir(osp.join(root_dir, SynthiaPath.SEMANTIC_SEGM_DIR)):
            gt_dir = osp.join(root_dir, SynthiaPath.SEMANTIC_SEGM_DIR)
            gt_images = find_images(gt_dir, recursive=True)
            for gt_img in gt_images:
                item_id = osp.splitext(osp.relpath(gt_img, gt_dir))[0].replace("\\", "/")

                anno = []
                inverse_cls_colormap = self._categories[AnnotationType.mask].inverse_colormap
                color_mask = lazy_mask(gt_img, inverse_cls_colormap)
                color_mask = color_mask()
                classes = np.unique(color_mask)
                for label_id in classes:
                    anno.append(
                        Mask(image=self._lazy_extract_mask(color_mask, label_id), label=label_id)
                    )

                image = images.get(item_id)
                if image:
                    image = Image(path=image)

                items[item_id] = DatasetItem(id=item_id, media=image, annotations=anno)

        return items

    @staticmethod
    def _lazy_extract_mask(mask, c):
        return lambda: mask == c


class SynthiaImporter(Importer):
    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        with context.require_any():
            for prefix in (
                SynthiaPath.IMAGES_DIR,
                SynthiaPath.LABELS_SEGM_DIR,
                SynthiaPath.SEMANTIC_SEGM_DIR,
            ):
                with context.alternative():
                    context.require_file(f"{prefix}/**/*.png")

    @classmethod
    def find_sources(cls, path):
        return [{"url": path, "format": "synthia"}]



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import contextlib
import csv
import fnmatch
import functools
import glob
import itertools
import json
import logging as log
import os
import os.path as osp
import re
import types

import cv2
import numpy as np
from attr import attrs

from datumaro.components.annotation import AnnotationType, Bbox, Label, LabelCategories, Mask
from datumaro.components.dataset import ItemStatus
from datumaro.components.dataset_base import DatasetBase, DatasetItem
from datumaro.components.errors import (
    DatasetError,
    MediaTypeError,
    RepeatedItemError,
    UndefinedLabel,
)
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.components.media import Image
from datumaro.components.validator import Severity
from datumaro.util import parse_json_file
from datumaro.util.annotation_util import find_instances
from datumaro.util.image import (
    DEFAULT_IMAGE_META_FILE_NAME,
    find_images,
    lazy_image,
    load_image,
    load_image_meta_file,
    save_image,
    save_image_meta_file,
)
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file
from datumaro.util.os_util import make_file_name, split_path

# A regex to check whether a string can be used as a "normal" path
# component.
# Some strings that are loaded from annotation files are used in paths,
# and we need to validate them with this, so that a specially crafted
# annotation file can't access files outside of the expected directory.
_RE_INVALID_PATH_COMPONENT = re.compile(
    r"""
    # empty
    | \.\.? # special path component
    | .*[/\\\0].* # contains special characters
""",
    re.VERBOSE,
)


@attrs(auto_attribs=True)
class UnsupportedSubsetNameError(DatasetError):
    item_id: str
    subset: str

    def __str__(self):
        return "Item %s has an unsupported subset name %r." % (self.item_id, self.subset)


@attrs(auto_attribs=True)
class UnsupportedBoxIdError(DatasetError):
    item_id: str
    box_id: str

    def __str__(self):
        return "Item %s has a mask with an unsupported box ID %r." % (self.item_id, self.box_id)


@attrs(auto_attribs=True)
class UnsupportedMaskPathError(DatasetError):
    item_id: str
    mask_path: str

    def __str__(self):
        return "Item %s has a mask with an unsupported path %r." % (self.item_id, self.mask_path)


class OpenImagesPath:
    ANNOTATIONS_DIR = "annotations"
    IMAGES_DIR = "images"
    MASKS_DIR = "masks"

    FULL_IMAGE_DESCRIPTION_FILE_NAME = "image_ids_and_rotation.csv"
    SUBSET_IMAGE_DESCRIPTION_FILE_PATTERNS = (
        "*-images-with-rotation.csv",
        "*-images-with-labels-with-rotation.csv",
    )
    V5_CLASS_DESCRIPTION_FILE_NAME = "class-descriptions.csv"
    V5_CLASS_DESCRIPTION_BBOX_FILE_NAME = "class-descriptions-boxable.csv"
    HIERARCHY_FILE_NAME = "bbox_labels_600_hierarchy.json"

    LABEL_DESCRIPTION_FILE_SUFFIX = "-annotations-human-imagelabels.csv"
    BBOX_DESCRIPTION_FILE_SUFFIX = "-annotations-bbox.csv"
    MASK_DESCRIPTION_FILE_SUFFIX = "-annotations-object-segmentation.csv"

    IMAGE_DESCRIPTION_FIELDS = (
        "ImageID",
        "Subset",
        "OriginalURL",
        "OriginalLandingURL",
        "License",
        "AuthorProfileURL",
        "Author",
        "Title",
        "OriginalSize",
        "OriginalMD5",
        "Thumbnail300KURL",
        "Rotation",
    )

    LABEL_DESCRIPTION_FIELDS = (
        "ImageID",
        "Source",
        "LabelName",
        "Confidence",
    )

    BBOX_DESCRIPTION_FIELDS = (
        "ImageID",
        "Source",
        "LabelName",
        "Confidence",
        "XMin",
        "XMax",
        "YMin",
        "YMax",
        "IsOccluded",
        "IsTruncated",
        "IsGroupOf",
        "IsDepiction",
        "IsInside",
    )

    BBOX_BOOLEAN_ATTRIBUTES = (
        types.SimpleNamespace(datumaro_name="occluded", oid_name="IsOccluded"),
        types.SimpleNamespace(datumaro_name="truncated", oid_name="IsTruncated"),
        types.SimpleNamespace(datumaro_name="is_group_of", oid_name="IsGroupOf"),
        types.SimpleNamespace(datumaro_name="is_depiction", oid_name="IsDepiction"),
        types.SimpleNamespace(datumaro_name="is_inside", oid_name="IsInside"),
    )

    MASK_DESCRIPTION_FIELDS = (
        "MaskPath",
        "ImageID",
        "LabelName",
        "BoxID",
        "BoxXMin",
        "BoxXMax",
        "BoxYMin",
        "BoxYMax",
        "PredictedIoU",
        "Clicks",
    )


class OpenImagesBase(DatasetBase):
    def __init__(self, path, image_meta=None):
        if not osp.isdir(path):
            raise FileNotFoundError("Can't read dataset directory '%s'" % path)

        super().__init__()

        self._dataset_dir = path

        self._annotation_dir = osp.join(path, OpenImagesPath.ANNOTATIONS_DIR)
        if not osp.exists(self._annotation_dir):
            self._annotation_dir = path
        self._annotation_files = os.listdir(self._annotation_dir)

        self._categories = {}
        self._items = []

        assert image_meta is None or isinstance(image_meta, (dict, str))
        if isinstance(image_meta, dict):
            self._image_meta = dict(image_meta)
        elif isinstance(image_meta, str):
            self._image_meta = load_image_meta_file(osp.join(path, image_meta))
        elif image_meta is None:
            try:
                self._image_meta = load_image_meta_file(
                    osp.join(self._annotation_dir, DEFAULT_IMAGE_META_FILE_NAME)
                )
            except FileNotFoundError:
                self._image_meta = {}

        if has_meta_file(path):
            self._categories = {
                AnnotationType.label: LabelCategories.from_iterable(parse_meta_file(path).keys())
            }
        else:
            self._load_categories()
        self._load_items()

    def __iter__(self):
        yield from self._items

    def categories(self):
        return self._categories

    @contextlib.contextmanager
    def _open_csv_annotation(self, file_name):
        absolute_path = osp.join(self._annotation_dir, file_name)

        with open(absolute_path, "r", encoding="utf-8", newline="") as f:
            yield csv.DictReader(f)

    def _glob_annotations(self, pattern):
        for annotation_file in self._annotation_files:
            if fnmatch.fnmatch(annotation_file, pattern):
                yield annotation_file

    def _load_categories(self):
        label_categories = LabelCategories()

        class_desc_patterns = [
            "oidv*-class-descriptions.csv",
            OpenImagesPath.V5_CLASS_DESCRIPTION_BBOX_FILE_NAME,
            OpenImagesPath.V5_CLASS_DESCRIPTION_FILE_NAME,
        ]

        class_desc_files = [
            file for pattern in class_desc_patterns for file in self._glob_annotations(pattern)
        ]

        if not class_desc_files:
            raise FileNotFoundError(
                "Can't find class description file, the "
                "annotations directory does't contain any of these files: %s"
                % ", ".join(class_desc_patterns)
            )

        # In OID v6, the class description file is prefixed with `oidv6-`, whereas
        # in the previous versions, it isn't. We try to find it regardless.
        # We use a wildcard so that if, say, OID v7 is released in the future with
        # a similar layout as v6, it's automatically supported.
        # If the file doesn't exist with either name, we'll fail trying to open
        # `class-descriptions.csv`.

        annotation_name = class_desc_files[0]
        with self._open_csv_annotation(annotation_name) as class_description_reader:
            # Prior to OID v6, this file didn't contain a header row.
            if annotation_name in {
                OpenImagesPath.V5_CLASS_DESCRIPTION_BBOX_FILE_NAME,
                OpenImagesPath.V5_CLASS_DESCRIPTION_FILE_NAME,
            }:
                class_description_reader.fieldnames = ("LabelName", "DisplayName")

            for class_description in class_description_reader:
                label_name = class_description["LabelName"]
                label_categories.add(label_name)

        self._categories[AnnotationType.label] = label_categories

        self._load_label_category_parents()

    def _load_label_category_parents(self):
        label_categories = self._categories[AnnotationType.label]

        hierarchy_path = osp.join(self._annotation_dir, OpenImagesPath.HIERARCHY_FILE_NAME)

        try:
            root_node = parse_json_file(hierarchy_path)
        except FileNotFoundError:
            return

        def set_parents_from_node(node, category):
            for child_node in node.get("Subcategory", []):
                _, child_category = label_categories.find(child_node["LabelName"])

                if category is not None and child_category is not None:
                    child_category.parent = category.name

                set_parents_from_node(child_node, child_category)

        _, root_category = label_categories.find(root_node["LabelName"])
        set_parents_from_node(root_node, root_category)

    def _load_items(self):
        images_dir = osp.join(self._dataset_dir, OpenImagesPath.IMAGES_DIR)

        self._image_paths_by_id = {
            # the first component of `path_parts` is the subset name
            "/".join(path_parts[1:]): path
            for path in find_images(images_dir, recursive=True)
            for path_parts in [
                split_path(
                    osp.splitext(osp.relpath(path, images_dir))[0],
                )
            ]
            if 1 < len(path_parts)
        }

        items_by_id = {}

        def load_from(annotation_name):
            with self._open_csv_annotation(annotation_name) as image_reader:
                for image_description in image_reader:
                    image_id = image_description["ImageID"]
                    if image_id in items_by_id:
                        raise RepeatedItemError(item_id=image_id)

                    subset = image_description["Subset"]

                    if _RE_INVALID_PATH_COMPONENT.fullmatch(subset):
                        raise UnsupportedSubsetNameError(item_id=image_id, subset=subset)

                    if image_id in items_by_id:
                        log.warning("Item %s is repeated" % image_id)
                        continue

                    items_by_id[image_id] = self._add_item(image_id, subset)

        # It's preferable to load the combined image description file,
        # because it contains descriptions for training images without human-annotated labels
        # (the file specific to the training set doesn't).
        # However, if it's missing, we'll try loading subset-specific files instead, so that
        # this extractor can be used on individual subsets of the dataset.
        try:
            load_from(OpenImagesPath.FULL_IMAGE_DESCRIPTION_FILE_NAME)
        except FileNotFoundError:
            for pattern in OpenImagesPath.SUBSET_IMAGE_DESCRIPTION_FILE_PATTERNS:
                for path in self._glob_annotations(pattern):
                    load_from(path)

        self._load_labels(items_by_id)
        normalized_coords = self._load_bboxes(items_by_id)
        self._load_masks(items_by_id, normalized_coords)

    def _add_item(self, item_id, subset):
        image_path = self._image_paths_by_id.get(item_id)
        image = None
        if image_path is None:
            log.warning(
                "Can't find image for item: %s. "
                "It should be in the '%s' directory" % (item_id, OpenImagesPath.IMAGES_DIR)
            )
        else:
            image = Image(path=image_path, size=self._image_meta.get(item_id))

        item = DatasetItem(id=item_id, media=image, subset=subset)
        self._items.append(item)
        return item

    def _get_subset_name(self, filename):
        parts = filename.split("-")
        return parts[1] if parts[0] == "oidv6" else parts[0]

    def _load_labels(self, items_by_id):
        label_categories = self._categories[AnnotationType.label]

        # TODO: implement reading of machine-annotated labels

        for label_path in self._glob_annotations(
            "*" + OpenImagesPath.LABEL_DESCRIPTION_FILE_SUFFIX
        ):
            with self._open_csv_annotation(label_path) as label_reader:
                for label_description in label_reader:
                    image_id = label_description["ImageID"]
                    item = items_by_id.get(image_id)
                    if item is None:
                        item = items_by_id.setdefault(
                            image_id, self._add_item(image_id, self._get_subset_name(label_path))
                        )

                    confidence = float(label_description["Confidence"])

                    label_name = label_description["LabelName"]
                    label_index, _ = label_categories.find(label_name)
                    if label_index is None:
                        raise UndefinedLabel(
                            item_id=item.id,
                            subset=item.subset,
                            label_name=label_name,
                            severity=Severity.error,
                        )
                    item.annotations.append(
                        Label(label=label_index, attributes={"score": confidence})
                    )

    def _load_bboxes(self, items_by_id):
        label_categories = self._categories[AnnotationType.label]

        # OID specifies box coordinates in the normalized form, which we have to
        # convert to the unnormalized form to fit the Datumaro data model.
        # However, we need to temporarily preserve the normalized form as well,
        # because we will need it later to match up box and mask annotations.
        # So we store each box's normalized coordinates in this dictionary.
        normalized_coords = {}

        for bbox_path in self._glob_annotations("*" + OpenImagesPath.BBOX_DESCRIPTION_FILE_SUFFIX):
            with self._open_csv_annotation(bbox_path) as bbox_reader:
                for bbox_description in bbox_reader:
                    image_id = bbox_description["ImageID"]
                    item = items_by_id.get(image_id)
                    if item is None:
                        item = items_by_id.setdefault(
                            image_id, self._add_item(image_id, self._get_subset_name(bbox_path))
                        )

                    label_name = bbox_description["LabelName"]
                    label_index, _ = label_categories.find(label_name)
                    if label_index is None:
                        raise UndefinedLabel(
                            item_id=item.id,
                            subset=item.subset,
                            label_name=label_name,
                            severity=Severity.error,
                        )

                    if item.media and item.media.size is not None:
                        height, width = item.media.size
                    elif self._image_meta.get(item.id):
                        height, width = self._image_meta[item.id]
                    else:
                        log.warning(
                            "Can't decode box for item '%s' due to missing image file", item.id
                        )
                        continue

                    x_min_norm, x_max_norm, y_min_norm, y_max_norm = [
                        float(bbox_description[field]) for field in ["XMin", "XMax", "YMin", "YMax"]
                    ]

                    x_min = x_min_norm * width
                    x_max = x_max_norm * width
                    y_min = y_min_norm * height
                    y_max = y_max_norm * height

                    attributes = {
                        "score": float(bbox_description["Confidence"]),
                    }

                    for bool_attr in OpenImagesPath.BBOX_BOOLEAN_ATTRIBUTES:
                        int_value = int(bbox_description[bool_attr.oid_name])
                        if int_value >= 0:
                            attributes[bool_attr.datumaro_name] = bool(int_value)

                    # Give each box within an item a distinct group ID,
                    # so that we can later group them together with the corresponding masks.
                    if item.annotations and item.annotations[-1].type is AnnotationType.bbox:
                        group = item.annotations[-1].group + 1
                    else:
                        group = 1

                    item.annotations.append(
                        Bbox(
                            label=label_index,
                            x=x_min,
                            y=y_min,
                            w=x_max - x_min,
                            h=y_max - y_min,
                            attributes=attributes,
                            group=group,
                        )
                    )

                    normalized_coords[id(item.annotations[-1])] = np.array(
                        [x_min_norm, x_max_norm, y_min_norm, y_max_norm]
                    )

        return normalized_coords

    def _load_masks(self, items_by_id, normalized_coords):
        label_categories = self._categories[AnnotationType.label]

        for mask_path in self._glob_annotations("*" + OpenImagesPath.MASK_DESCRIPTION_FILE_SUFFIX):
            with self._open_csv_annotation(mask_path) as mask_reader:
                for mask_description in mask_reader:
                    mask_path = mask_description["MaskPath"]
                    if _RE_INVALID_PATH_COMPONENT.fullmatch(mask_path):
                        raise UnsupportedMaskPathError(item_id=item.id, mask_path=mask_path)

                    image_id = mask_description["ImageID"]
                    item = items_by_id.get(image_id)
                    if item is None:
                        item = items_by_id.setdefault(
                            image_id, self._add_item(image_id, self._get_subset_name(mask_path))
                        )

                    label_name = mask_description["LabelName"]
                    label_index, _ = label_categories.find(label_name)
                    if label_index is None:
                        raise UndefinedLabel(
                            item_id=item.id,
                            subset=item.subset,
                            label_name=label_name,
                            severity=Severity.error,
                        )

                    if item.media and item.media.has_size:
                        image_size = item.media.size
                    elif self._image_meta.get(item.id):
                        image_size = self._image_meta.get(item.id)
                    else:
                        log.warning(
                            "Can't decode mask for item '%s' due to missing image file", item.id
                        )
                        continue

                    attributes = {}

                    # The box IDs are rather useless, because the _box_ annotations
                    # don't include them, so they cannot be used to match masks to boxes.
                    # However, it is still desirable to record them, because they are
                    # included in the mask file names, so in order to save each mask to the
                    # file it was loaded from when saving in-places, we need to know
                    # the original box ID.
                    box_id = mask_description["BoxID"]
                    if _RE_INVALID_PATH_COMPONENT.fullmatch(box_id):
                        raise UnsupportedBoxIdError(item_id=item.id, box_id=box_id)
                    attributes["box_id"] = box_id

                    group = 0

                    box_coord_fields = ("BoxXMin", "BoxXMax", "BoxYMin", "BoxYMax")

                    # The original OID has box coordinates for all masks, but
                    # a dataset converted from another dataset might not.
                    if all(mask_description[f] for f in box_coord_fields):
                        # Try to find the box annotation corresponding to the
                        # current mask.
                        mask_box_coords = np.array(
                            [float(mask_description[field]) for field in box_coord_fields]
                        )

                        for annotation in item.annotations:
                            if (
                                annotation.type is AnnotationType.bbox
                                and annotation.label == label_index
                            ):
                                # In the original OID, mask box coordinates are stored
                                # with 6 digit precision, hence the tolerance.
                                if np.allclose(
                                    mask_box_coords,
                                    normalized_coords[id(annotation)],
                                    rtol=0,
                                    atol=1e-6,
                                ):
                                    group = annotation.group

                    if mask_description["PredictedIoU"]:
                        attributes["predicted_iou"] = float(mask_description["PredictedIoU"])

                    item.annotations.append(
                        Mask(
                            image=lazy_image(
                                osp.join(
                                    self._dataset_dir,
                                    OpenImagesPath.MASKS_DIR,
                                    item.subset,
                                    mask_path,
                                ),
                                loader=functools.partial(
                                    self._load_and_resize_mask, size=image_size
                                ),
                            ),
                            label=label_index,
                            attributes=attributes,
                            group=group,
                        )
                    )

    @staticmethod
    def _load_and_resize_mask(path, size):
        raw = load_image(path, dtype=np.uint8)
        resized = cv2.resize(raw, (size[1], size[0]), interpolation=cv2.INTER_NEAREST)
        return resized.astype(bool)


class OpenImagesImporter(Importer):
    POSSIBLE_ANNOTATION_PATTERNS = (
        OpenImagesPath.FULL_IMAGE_DESCRIPTION_FILE_NAME,
        *OpenImagesPath.SUBSET_IMAGE_DESCRIPTION_FILE_PATTERNS,
        "*" + OpenImagesPath.LABEL_DESCRIPTION_FILE_SUFFIX,
        "*" + OpenImagesPath.BBOX_DESCRIPTION_FILE_SUFFIX,
        "*" + OpenImagesPath.MASK_DESCRIPTION_FILE_SUFFIX,
    )

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        ann_dirs = [f"{OpenImagesPath.ANNOTATIONS_DIR}/", ""]
        ann_patterns = itertools.product(ann_dirs, cls.POSSIBLE_ANNOTATION_PATTERNS)
        with context.require_any():
            for ann_dir, ann_file in ann_patterns:
                with context.alternative():
                    context.require_file(ann_dir + ann_file)

    @classmethod
    def find_sources(cls, path):
        for pattern in cls.POSSIBLE_ANNOTATION_PATTERNS:
            if glob.glob(osp.join(glob.escape(path), OpenImagesPath.ANNOTATIONS_DIR, pattern)):
                return [{"url": path, "format": OpenImagesBase.NAME}]
            elif glob.glob(osp.join(glob.escape(path), pattern)):
                return [{"url": path, "format": OpenImagesBase.NAME}]
        return []


class _LazyCsvDictWriter:
    """
    For annotation files, we only learn that the file is required after
    we find at least one occurrence of the corresponding annotation type.
    However, it's convenient to create the writer ahead of time, so that
    it can be used in a `with` statement.

    This class behaves like a csv.DictWriter, but it only creates the file
    once the first row is written.
    """

    def __init__(self, writer_manager_factory):
        self._writer_manager_factory = writer_manager_factory
        self._writer_manager = None
        self._writer = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._writer_manager:
            self._writer_manager.__exit__(exc_type, exc_value, traceback)

    def writerow(self, rowdict):
        if not self._writer_manager:
            self._writer_manager = self._writer_manager_factory()
            self._writer = self._writer_manager.__enter__()

        self._writer.writerow(rowdict)


class _AnnotationWriter:
    _POSSIBLE_ANNOTATION_FILE_PATTERNS = (
        # class description files don't need to be listed,
        # because they are always written
        OpenImagesPath.FULL_IMAGE_DESCRIPTION_FILE_NAME,
        *OpenImagesPath.SUBSET_IMAGE_DESCRIPTION_FILE_PATTERNS,
        "*" + OpenImagesPath.LABEL_DESCRIPTION_FILE_SUFFIX,
        "*" + OpenImagesPath.BBOX_DESCRIPTION_FILE_SUFFIX,
        "*" + OpenImagesPath.MASK_DESCRIPTION_FILE_SUFFIX,
        DEFAULT_IMAGE_META_FILE_NAME,
    )

    def __init__(self, root_dir):
        self._annotations_dir = osp.join(root_dir, OpenImagesPath.ANNOTATIONS_DIR)
        self._written_annotations = set()

        os.makedirs(self._annotations_dir, exist_ok=True)

    @contextlib.contextmanager
    def open(self, file_name, newline=None):
        self._written_annotations.add(file_name)

        file_path = osp.join(self._annotations_dir, file_name)

        # Write to a temporary file first, to avoid data loss if we're patching
        # an existing dataset and the process is interrupted.
        temp_file_path = file_path + ".tmp"

        with open(
            temp_file_path,
            "w",
            encoding="utf-8",
            newline=newline,
        ) as f:
            yield f

        os.replace(temp_file_path, file_path)

    @contextlib.contextmanager
    def open_csv(self, file_name, field_names, *, write_header=True):
        with self.open(file_name, newline="") as f:
            writer = csv.DictWriter(f, field_names)
            if write_header:
                writer.writeheader()
            yield writer

    def open_csv_lazy(self, file_name, field_names):
        return _LazyCsvDictWriter(lambda: self.open_csv(file_name, field_names))

    def remove_unwritten(self):
        for file_name in os.listdir(self._annotations_dir):
            if file_name not in self._written_annotations and any(
                fnmatch.fnmatch(file_name, pattern)
                for pattern in self._POSSIBLE_ANNOTATION_FILE_PATTERNS
            ):
                os.unlink(osp.join(self._annotations_dir, file_name))


class OpenImagesExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".jpg"

    def apply(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        self._save(_AnnotationWriter(self._save_dir))

    @classmethod
    def patch(cls, dataset, patch, save_dir, **options):
        converter = cls(dataset, save_dir, **options)
        annotation_writer = _AnnotationWriter(save_dir)
        converter._save(annotation_writer)
        annotation_writer.remove_unwritten()

        images_dir = osp.join(save_dir, OpenImagesPath.IMAGES_DIR)
        for (item_id, subset), status in patch.updated_items.items():
            if status != ItemStatus.removed:
                continue

            item = DatasetItem(item_id, subset=subset)

            image_path = osp.join(images_dir, converter._make_image_filename(item, subdir=subset))

            if osp.isfile(image_path):
                os.unlink(image_path)

    def _save(self, annotation_writer):
        self._save_categories(annotation_writer)
        self._save_label_category_parents(annotation_writer)
        self._save_subsets(annotation_writer)

    def _save_categories(self, annotation_writer):
        if self._save_dataset_meta:
            self._save_meta_file(self._save_dir)

        with annotation_writer.open_csv(
            OpenImagesPath.V5_CLASS_DESCRIPTION_FILE_NAME,
            ["LabelName", "DisplayName"],
            # no header, since we're saving it in the V5 format
            write_header=False,
        ) as class_description_writer:
            for category in self._extractor.categories()[AnnotationType.label]:
                class_description_writer.writerow(
                    {
                        "LabelName": category.name,
                        "DisplayName": category.name,
                    }
                )

    def _save_label_category_parents(self, annotation_writer):
        all_label_names = set()
        hierarchy_nodes = {}
        orphan_nodes = []

        def get_node(name):
            return hierarchy_nodes.setdefault(name, {"LabelName": name})

        for category in self._extractor.categories()[AnnotationType.label]:
            all_label_names.add(category.name)

            child_node = get_node(category.name)

            if category.parent:
                parent_node = get_node(category.parent)
                parent_node.setdefault("Subcategory", []).append(child_node)
            else:
                orphan_nodes.append(child_node)

        # The hierarchy has to be rooted in a single node. However, there's
        # no guarantee that there exists only one orphan (label without a parent).
        # Therefore, we create a fake root node and make it the parent of every
        # orphan label.
        # This is not a violation of the format, because the original OID does
        # the same thing.
        root_node = {
            # Create an OID-like label name that isn't already used by a real label
            "LabelName": next(
                root_name
                for i in itertools.count()
                for root_name in [f"/m/{i}"]
                if root_name not in all_label_names
            ),
            # If an orphan has no children, then it makes no semantic difference
            # whether it's listed in the hierarchy file or not. So strip such nodes
            # to avoid recording meaningless data.
            "Subcategory": [node for node in orphan_nodes if "Subcategory" in node],
        }

        hierarchy_path = osp.join(
            self._save_dir, OpenImagesPath.ANNOTATIONS_DIR, OpenImagesPath.HIERARCHY_FILE_NAME
        )

        with annotation_writer.open(hierarchy_path) as hierarchy_file:
            json.dump(root_node, hierarchy_file, indent=4, ensure_ascii=False)
            hierarchy_file.write("\n")

    def _save_subsets(self, annotation_writer):
        label_categories = self._extractor.categories().get(AnnotationType.label, LabelCategories())

        image_meta = {}

        for subset_name, subset in self._extractor.subsets().items():
            if _RE_INVALID_PATH_COMPONENT.fullmatch(subset_name):
                raise UnsupportedSubsetNameError(item_id=next(iter(subset)).id, subset=subset)

            image_description_name = f"{subset_name}-images-with-rotation.csv"

            with annotation_writer.open_csv(
                image_description_name,
                OpenImagesPath.IMAGE_DESCRIPTION_FIELDS,
            ) as image_description_writer, annotation_writer.open_csv_lazy(
                subset_name + OpenImagesPath.LABEL_DESCRIPTION_FILE_SUFFIX,
                OpenImagesPath.LABEL_DESCRIPTION_FIELDS,
            ) as label_description_writer, annotation_writer.open_csv_lazy(
                subset_name + OpenImagesPath.BBOX_DESCRIPTION_FILE_SUFFIX,
                OpenImagesPath.BBOX_DESCRIPTION_FIELDS,
            ) as bbox_description_writer, annotation_writer.open_csv_lazy(
                subset_name + OpenImagesPath.MASK_DESCRIPTION_FILE_SUFFIX,
                OpenImagesPath.MASK_DESCRIPTION_FIELDS,
            ) as mask_description_writer:
                for item in subset:
                    image_description_writer.writerow(
                        {
                            "ImageID": item.id,
                            "Subset": subset_name,
                        }
                    )

                    if self._save_media:
                        if item.media:
                            self._save_image(
                                item, subdir=osp.join(OpenImagesPath.IMAGES_DIR, subset_name)
                            )
                        else:
                            log.debug("Item '%s' has no image", item.id)

                    self._save_item_annotations(
                        item,
                        label_description_writer,
                        bbox_description_writer,
                        mask_description_writer,
                        label_categories,
                        image_meta,
                    )

        if image_meta:
            image_meta_file_path = osp.join(
                self._save_dir, OpenImagesPath.ANNOTATIONS_DIR, DEFAULT_IMAGE_META_FILE_NAME
            )

            save_image_meta_file(image_meta, image_meta_file_path)

    def _save_item_annotations(
        self,
        item,
        label_description_writer,
        bbox_description_writer,
        mask_description_writer,
        label_categories,
        image_meta,
    ):
        next_box_id = 0

        existing_box_ids = {
            annotation.attributes["box_id"]
            for annotation in item.annotations
            if annotation.type is AnnotationType.mask
            if "box_id" in annotation.attributes
        }

        for instance in find_instances(item.annotations):
            instance_box = next((a for a in instance if a.type is AnnotationType.bbox), None)

            for annotation in instance:
                if annotation.type is AnnotationType.label:
                    label_description_writer.writerow(
                        {
                            "ImageID": item.id,
                            "LabelName": label_categories[annotation.label].name,
                            "Confidence": str(annotation.attributes.get("score", 1)),
                        }
                    )
                elif annotation.type is AnnotationType.bbox:
                    if item.media and item.media.size is not None:
                        image_meta[item.id] = item.media.size
                        height, width = item.media.size
                    else:
                        log.warning(
                            "Can't encode box for item '%s' due to missing image file", item.id
                        )
                        continue

                    bbox_description_writer.writerow(
                        {
                            "ImageID": item.id,
                            "LabelName": label_categories[annotation.label].name,
                            "Confidence": str(annotation.attributes.get("score", 1)),
                            "XMin": annotation.x / width,
                            "YMin": annotation.y / height,
                            "XMax": (annotation.x + annotation.w) / width,
                            "YMax": (annotation.y + annotation.h) / height,
                            **{
                                bool_attr.oid_name: int(
                                    annotation.attributes.get(bool_attr.datumaro_name, -1)
                                )
                                for bool_attr in OpenImagesPath.BBOX_BOOLEAN_ATTRIBUTES
                            },
                        }
                    )
                elif annotation.type is AnnotationType.mask:
                    mask_dir = osp.join(self._save_dir, OpenImagesPath.MASKS_DIR, item.subset)

                    box_id_str = annotation.attributes.get("box_id")

                    if box_id_str:
                        if _RE_INVALID_PATH_COMPONENT.fullmatch(box_id_str):
                            raise UnsupportedBoxIdError(item_id=item.id, box_id=box_id_str)
                    else:
                        # find a box ID that isn't used in any other annotations
                        while True:
                            box_id_str = format(next_box_id, "08x")
                            next_box_id += 1
                            if box_id_str not in existing_box_ids:
                                break

                    label_name = label_categories[annotation.label].name
                    mask_file_name = "%s_%s_%s.png" % (
                        make_file_name(item.id),
                        make_file_name(label_name),
                        box_id_str,
                    )

                    box_coords = {}

                    if instance_box is not None:
                        if item.media and item.media.size is not None:
                            image_meta[item.id] = item.media.size
                            height, width = item.media.size

                            box_coords = {
                                "BoxXMin": instance_box.x / width,
                                "BoxXMax": (instance_box.x + instance_box.w) / width,
                                "BoxYMin": instance_box.y / height,
                                "BoxYMax": (instance_box.y + instance_box.h) / height,
                            }
                        else:
                            log.warning(
                                "Can't encode box coordinates for a mask"
                                " for item '%s' due to missing image file",
                                item.id,
                            )

                    mask_description_writer.writerow(
                        {
                            "MaskPath": mask_file_name,
                            "ImageID": item.id,
                            "LabelName": label_name,
                            "BoxID": box_id_str,
                            **box_coords,
                            "PredictedIoU": annotation.attributes.get("predicted_iou", ""),
                        }
                    )

                    save_image(
                        osp.join(mask_dir, mask_file_name), annotation.image, create_dir=True
                    )
